## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2019-2023 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2023-01-26
# @author Lukasz Butkowski  <lukasz.butkowski@desy.de>
# @author Michael Buechler <michael.buechler@desy.de>
# @author Seyed Nima Omidsajedi <nima.sajedi@desy.de>
# --------------------------------------------------------------------------- #
# @brief
# Part of DESY FPGA Firmware Framework (fwk)
# Main FPGA firmware framework with namespaces definitions
# --------------------------------------------------------------------------- #

# ==============================================================================
## main framework namespace
# consist of namespace variables and procedures
namespace eval ::fwk {

  # main framework variables
  variable ProjectPath

  variable ProjectName
  variable ProjectConf
  variable ToolType
  variable CoSimTool ""

  variable PrjPath
  variable PrjBuildName
  variable PrjBuildPath
  variable ReleasePath
  variable ReleaseName

  variable SrcPath
#  variable LibPath
  variable DefLibrary "deflibrary"

  variable NamespacesList {}
  variable ModNamespacesList {}

  variable VerMajor 0
  variable VerMinor 0
  variable VerPatch 0
  variable VerCommits 0
  variable VerShasum 0
  variable VerString ""

  variable IpFlow 0

  # array set AddressSpace {}
  # set AddressSpace   {}
  # set AddressSpaceDict   {}
  # set AddressSpaceChDict {}
  # set RdlFiles {}

  variable SimTop {}
  variable SimTime {}
  # ==============================================================================
  ## @brief
  # addFwModule {ns InstanceName TclFile}
  #   adds firmware module to the fwk namespace
  #   creates a new namespace under ::fwk::src::{InstanceName} with dedicated procedures
  #   as well sources provided module TclFile
  # @param ns            name of the namespace where module should be included
  # @param ModuleDict    module parameters in a form of dictionaray
  #
  proc addModule {ns ModuleDict} {}

  # ==============================================================================
  ## loadProjectInfo::
  #   loads project information and configuration form the provided config file cfgFile
  #
  proc loadProjectInfo {cfgFile} {}

  # ==============================================================================
  ## printProjectInfo:: prints to output main project settings
  #
  proc printProjectInfo {} {}


  # ==============================================================================
  ## setSources::
  #   executes addSources procedure from all namespacess
  #   recursively from last children to parent
  #
  proc setSources {{ns ::}} {}

  # ==============================================================================
  ## setProperties::
  #   executes addSources procedure from all namespacess
  #   recursively from last children to parent
  #
  proc setProperties {{ns ::}} {}

  # ==============================================================================
  ## addSources:: add sources to project in selected namespaces
  #   it uses ToolType variable to run tool depended procedure
  # Argument:
  #   namespaces - namespace where the list of sources is located
  #   args - variable name with sources list and additional arguments
  #          to pass to the tool procedure like "-fileset test"
  #
  proc addSources {ns} {}

  # ==============================================================================
  ## @brief
  # getVersions::
  # based on git tag sets VerMajor VerMinor VerPatch VerCsh VerCommits
  proc getVersion {{ns ::}} {}

  proc genPrjVerFile {{ns} fileType fileName} {}
  proc genModVerFile {{ns} fileType fileName} {}
  # proc addAddressSpace {{ns} {_AddressSpace} Name Type BaseAddress Args {Config {}}} {}

  # ==============================================================================
  ##   create a list of available namespaces under ::fwk::NamespacesList
  #   default namespace ::fwk
  #
  proc createNamespacesList {{ns ::fwk}} {}
  # ==============================================================================
  ## setSwSources::
  #   executes swAddSources procedure from all namespacess
  #   recursively from last children to parent
  #
  proc setSwSources {name ipDir} {
    variable sw_src_name
    sourceFwkFile $ipDir
    lappend sw_src_name $name
  }

  # ==============================================================================
  ## @brief
  # genHdlFromConfig {}
  #  Exports the contents of Config variable as VHDL package using Jinja tool
  # and adds it to the Module's sources variable
  proc genHdlFromConfig {} {}

}

# ==============================================================================
# ==============================================================================
# Make procedures visible
namespace eval ::fwk {
  namespace export doOnCreate \
  setSources \
  addModule \
  addSrcModuleAsIp \
  printProjectInfo \
  parseVhdlConfigFile \
  setProperties \
  addSources \
  getVersion \
  genPrjVerFile \
  genModVerFile \
  setSwSources \
  sw_src_dir \
  sw_src_name \
  listSubmodules \
  getSubmoduleVersion \
  genHdlFromConfig

}
# globally visible procedures which are under each namespace
proc addModule {InstanceName TclFile} {}
proc addSrcModuleAsIp {InstanceName TclFile} {}
proc addSources {args} {}
proc parseVhdlConfigFile {varName vhdlFile} {}
proc getVersion {} {}
proc genModVerFile {fileType fileName} {}
proc genPrjVerFile {fileType fileName} {}
# proc addAddressSpace {{_AddressSpace} Name Type BaseAddress Args {Config {}}} {}

# ==============================================================================
# ==============================================================================
proc ::fwk::addModule {ns ModuleDict} {

  #check keys
  set InstanceName [dict get $ModuleDict InstanceName]
  set TclFile [dict get $ModuleDict File]

  ::fwk::printInfo "Creating namespace ${ns}::${InstanceName}"

  if { [namespace exists ${ns}::${InstanceName}] } {
    ::fwk::printError "namespace ${ns}::${InstanceName} already exists, please add Module with instance name: ${InstanceName} only ones."
    ::fwk::exit -2
  }

  lappend ::fwk::SrcNamespacesList ${ns}::${InstanceName}

  # check if Tcl file exist before creating new namespace
  # set tclfile to temp to be accessible for created namespace
  if { [ file isfile $TclFile ] }  {
    # full path
    set ::fwk::tmpTclPath [file normalize $TclFile ]
  } elseif { [ file isfile $::fwk::ProjectPath/$TclFile ] } {
    # relative to project path
    set ::fwk::tmpTclPath [file normalize $::fwk::ProjectPath/$TclFile ]
  } else {
    # relative to tcl file
    set curDir [pwd]
    if {[info exists ${ns}::TclPath] } {
      cd [subst $${ns}::TclPath]
    }
    if { [ file isfile $TclFile ] } {
      set ::fwk::tmpTclPath [file normalize $TclFile ]
    } else {
      ::fwk::printError "${ns}::${InstanceName} cannot open Tcl file, namespace not created\n"
      ::fwk::exit -2
    }
    cd $curDir
  }

  # Save the Configuration parameters inside fwk namespace 
  # such that we can influence the ${ns}::${InstanceName} namespace
  # set ::fwk::tmpModConfVars $ConfigVars
  set ::fwk::tmpModuleDict $ModuleDict

  # create namespace for the module
  namespace eval ${ns}::${InstanceName} {
    variable IsSrcModule 1

    variable Name ""
    variable Info
    variable Config
    variable Sources {}
    variable Templates {}

    variable Path ""
    variable CurrentNamespace {}

    variable VerMajor 0
    variable VerMinor 0
    variable VerPatch 0
    variable VerCommits 0
    variable VerShasum 0
    variable VerString ""
    variable VerHex 00000000
    variable Ver
    variable Timestamp

    # array set AddressSpace {}
    variable AddressSpace {}

    set CurrentNamespace [namespace current]
    set InstanceName [namespace tail $CurrentNamespace]
    set NsLevel [expr {[llength [split $CurrentNamespace "::"]]/2 - 1}]

    # aliases to ::fwk function to be available in new namespace
    interp alias {}  ${CurrentNamespace}::addModule {} ::fwk::addModule ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::addSource {} ::fwk::addSource ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::addSources {} ::fwk::addSources ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::procTemplates {} ::fwk::procTemplates ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::addFwkIp {} ::fwk::addFwkIp
    interp alias {}  ${CurrentNamespace}::addGenIPSources {} ::fwk::addGenIPSources ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::parseVhdlConfigFile {} ::fwk::parseVhdlConfigFile ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::getVersion {} ::fwk::getVersion ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::genModVerFile {} ::fwk::genModVerFile ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::genPrjVerFile {} ::fwk::genPrjVerFile ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::addAddressSpace {} ::fwk::addAddressSpace ${CurrentNamespace}
    interp alias {}  ${CurrentNamespace}::genHdlFromConfig {} ::fwk::genHdlFromConfig ${CurrentNamespace}

    # sourceFwkFile module tcl file, check absolute and relative path
    set TclFile $::fwk::tmpTclPath
    set TclPath [file dirname [file normalize $TclFile]]
    set Path [file normalize $TclPath/.. ]

    sourceFwkFile $TclFile

    # run init procedure in namespace if exists
    set curDir [pwd]
    cd [subst $${CurrentNamespace}::TclPath]

    set Ver 0
    set Timestamp 0
    # get module version
    ::fwk::getVersion $CurrentNamespace

    # inject versions to configuration variable
    set Config(C_VERSION)       [format 0x%08x $Ver]
    set Config(C_TIMESTAMP)     $Timestamp
    set Config(C_PRJ_VERSION)   [format 0x%08x $::fwk::Ver]
    set Config(C_PRJ_SHASUM)    $::fwk::VerShasum
    set Config(C_PRJ_TIMESTAMP) $::fwk::Timestamp

    # Inject the Config parameters inside the namespace BEFORE 
    # running the init proc of a module in order to make sure
    # higher level namespace can influence the init Config parameters.
    if { [llength [namespace which ${CurrentNamespace}::doOnInit]] } {
      if { [dict exists $fwk::tmpModuleDict Config]} {
        foreach {key value} [dict get $::fwk::tmpModuleDict Config] {
          set Config($key) $value
        }
      }
      ::fwk::printInfo "executing ${CurrentNamespace}::doOnInit"
      ::fwk::catchHdl ${CurrentNamespace}::doOnInit
    } else {
      ::fwk::printWarning "$CurrentNamespace has no doOnInit procedure"
    }
    cd $curDir
  }

  unset -nocomplain ::fwk::tmpModuleDict
  unset -nocomplain ::fwk::tmpTclPath

}

# ==============================================================================
# generate module in namespace as an IP using tool dependent functions
proc ::fwk::packageIp {ns} {
  set InstanceName [subst $${ns}::ModuleName]

  set ::fwk::IpFlow 1

  ::fwk::printCBM "\n== CREATE/PACKAGE IP  $ModuleName =="

  set curDir [pwd]
  file mkdir ${::fwk::PrjPath}/ip_${InstanceName}
  cd $::fwk::PrjPath/ip_${InstanceName}

  ::fwk::tool::createProject ip_$ModuleName

  #
  # follow a similar flow to ::fwk::createProject
  #
  if { $::fwk::addr::TypesToGen != {} } {
    ::fwk::addr::main $::fwk::addr::TypesToGen ::fwk::ip::${InstanceName}::AddressSpace ip_${InstanceName}
    ::fwk::addr::addAddrSrc
  } else { ::fwk::printInfo "No address space outputs to generate set in ::fwk::addr::TypesToGen" }

  ::fwk::setSources ::fwk::ip::$ModuleName
  # extra steps, e.g. generate address space sources and mapfile

  ::fwk::doOnCreate ::fwk::ip::$ModuleName

  # call tool-specific IP packaging procedure
  ::fwk::tool::packageIp $::fwk::IpRepoPath ::fwk::ip::$ModuleName

  ::fwk::tool::closeProject

  # # don't repeat these as part of the main project flow
  # rename ${ns}::${InstanceName}::setSources ""
  # proc ${ns}::${InstanceName}::setSources {} {puts "empty setSources"}
  # rename ${ns}::${InstanceName}::doOnCreate ""
  # proc ${ns}::${InstanceName}::doOnCreate {} {puts "empty doOnCreate"}
  ::fwk::printSuccess "IP $ModuleName Sucessfully created\n"

  set ::fwk::IpFlow 0

  cd $curDir

}

# ==============================================================================
proc ::fwk::loadProjectInfo {cfgFile} {

  global ExecutionPath

  set configFile [file join $::fwk::ProjectPath cfg default.cfg]
  ::fwk::printInfo "Getting default project configuration from: default.cfg"

  if {[file exists $configFile]} {
    set Cfg [open $configFile r]
    while { [gets $Cfg line] >= 0 } {
      set match [regexp {^(?!#\n)(\w+)\s*=\s*(.*)$} $line totalmatch name value]
      if {1 == $match} {
        set $name  $value
        set ::fwk::$name $value
      }
    }

  } else  {
    ::fwk::printError "Cannot open default.cfg"
    ::fwk::exit -7
  }

  # check provided cfgFile
  if {[file exists $ExecutionPath/$cfgFile]} {
    # if provided in relation to execution folders
    set cfgFile $ExecutionPath/$cfgFile
  } elseif {[file exists $cfgFile]} {
    # if provided in absolute path
  } else {
    # file do not exist
      ::fwk::printError "$cfgFile does not exist"
      ::fwk::exit -7
  }

  # updating default configuration with changed config
  if { [file normalize $configFile] != [file normalize $cfgFile]  } {
      ::fwk::printInfo "Getting project configuration from: $cfgFile"

    set Cfg [open $cfgFile r]
    while { [gets $Cfg line] >= 0 } {
      set match [regexp {^(?!#\n)(\w+)\s*=\s*(.*)$} $line totalmatch name value]
      if {1 == $match} {
        set $name  $value
        set ::fwk::$name $value
      }
    }

  }

  set ::fwk::PrjBuildName ${::fwk::ProjectName}_${::fwk::ProjectConf}
  set ::fwk::PrjPath      [file join $::fwk::ProjectPath prj]
  set ::fwk::PrjBuildPath [file join $::fwk::PrjPath $::fwk::PrjBuildName]
  set ::fwk::PrjCmpPath   [file join $::fwk::PrjBuildPath compiled_res]
  set ::fwk::WorkspacePath $::fwk::PrjBuildPath
  set ::fwk::PlatformName  ${::fwk::ProjectName}_platform
  set ::fwk::IpRepoPath    ${::fwk::PrjBuildPath}/ip_repo
  set ::fwk::HlsIpRepoPath ${::fwk::PrjPath}/hls_ip_repo

  if { [info exists ::env(FWK_HW_FILE)] } {

    set hwFilePath [lindex [lsort -dictionary [glob -nocomplain $::env(FWK_HW_FILE)]] end]
    if { [file exists $hwFilePath] } {  # absolute path
      set ::fwk::HwFile $hwFilePath
    }

    set hwFilePath [lindex [lsort -dictionary [glob -nocomplain $ExecutionPath/$::env(FWK_HW_FILE)]] end]
    if { [file exists $hwFilePath] } {  # relative path
      set ::fwk::HwFile $hwFilePath
    }
    if { ![info exists ::fwk::HwFile] } {
      set ::fwk::HwFile "" ; #$::env(FWK_HW_FILE)
      ::fwk::printWarning "Cannot find HW File: $::env(FWK_HW_FILE)"
    }
  }

  set ::fwk::SrcPath ${::fwk::ProjectPath}/src

  # set library path
  if {![llength [namespace which -variable ::fwk::LibPath]]} {
    # if variable does not exists
    # set default lib path
    set ::fwk::LibPath ${::fwk::ProjectPath}/src/lib
    # if path does not exist
  } elseif {![file exists ${::fwk::ProjectPath}/${::fwk::LibPath} ]} {
    # set default lib path
    set ::fwk::LibPath ${::fwk::ProjectPath}/src/lib
  } else {
    set ::fwk::LibPath  ${::fwk::ProjectPath}/${::fwk::LibPath}
  }

}


# ==============================================================================
proc ::fwk::createNamespacesList {{ns ::}} {

  foreach ins [namespace children $ns] {
    ::fwk::createNamespacesList ${ins}
  }
  lappend ::fwk::NamespacesList $ns
}
# ==============================================================================
proc ::fwk::genSrcTree {} {
  ::fwk::printDivider
  set tree [list]
  lappend tree "Tree View of the Sources:"
  set nl [::fwk::createNamespacesList ::fwk::src]
  set snl [lsort $nl]
  foreach ele $snl {
    set depth [subst $${ele}::NsLevel]
    set curDepth [expr {$depth *2}]
    set curName [lindex [split ${ele} ::] end]
    # ::fwk::printInfo [format "%*s- %-*s" ${curDepth} "+"  24 ${curName}]
    lappend tree [format "%*s- %-*s" ${curDepth} "+"  24 ${curName}]
  }
  ::fwk::printInfoMLine $tree
}
# ==============================================================================
# #n
# proc ::fwk::genSwSrcTree  {} {
#   puts "----------------------------------------------------"
#   puts "List of SW IPs:"
#   foreach item $::fwk::sw_src_name {
#     puts "  + $item"
#   }
#   puts " --------------------"
#   puts $::fwk::sw_src_dir
#   puts "----------------------------------------------------"
# }
# #n

# ==============================================================================
# open project
proc ::fwk::openProject {} {
  ::fwk::printInfo "Openning $::fwk::PrjBuildName project ...\n"

  # change directory to the current project build
  cd $::fwk::PrjBuildPath

  ::fwk::tool::openProject $::fwk::PrjBuildName
}

# ==============================================================================
proc ::fwk::createProject {} {

  #check if project build folder exists, create prj and project build folder
  ::fwk::createPrjDir

  # change directory to the current project build
  cd $::fwk::PrjBuildPath

  # deletes existing project files if existing and recreates it
  ::fwk::tool::cleanProject $::fwk::PrjBuildName

  # package IPS
  if { [namespace exists ::fwk::ip]} {
    ::fwk::printDivider
    ::fwk::printCBM "== PACKAGE MODULES INTO VENDOR IPS =="
    ::fwk::iterateModules ::fwk::ip {::fwk::packageIp [namespace current]}
  }

  ::fwk::printDivider
  ::fwk::printCBM "== CREATE =="
  ::fwk::printInfo "Creating ( $::fwk::PrjBuildName ) $::fwk::ToolType project ..."
  ::fwk::printInfo "in $::fwk::PrjBuildPath"

  ::fwk::tool::createProject $::fwk::PrjBuildName

  # ::fwk::printDivider
  # ::fwk::printCBM "== SET PRJ PROPERTIES =="

  # # set project properties, FPGA type, etc..
  # MOVE TO DO ON CREATE
  # if { $::fwk::tool::SetPrjProperties } { ::fwk::setPrjProperties }

  ::fwk::printDivider
  ::fwk::printCBM "== DO ON CREATE =="
  ::fwk::doOnCreate ::fwk::src
  ::fwk::doOnCreate ::fwk::lib

  ::fwk::printDivider
  ::fwk::printCBM "== PROC TEMPLATES =="
  ::fwk::procTemplatesRec

  # Convert Config variables to HDL Package using Jinja2 templating engine
  ::fwk::genHdlFromConfig

  ::fwk::printDivider
  ::fwk::printCBM "== GEN/ADD ADDRESS SPACE =="
  if { $::fwk::addr::TypesToGen != {} } {
    ::fwk::addr::main $::fwk::addr::TypesToGen
    ::fwk::addr::addAddrSrc
    ::fwk::addSources ::fwk::addr
  } else { puts "Noting to generate set in ::fwk::addr::TypesToGen" }

  ::fwk::printDivider
  ::fwk::printCBM "== ADD SOURCES =="

  ::fwk::addSourcesRec


  ::fwk::printDivider
  ::fwk::printCBM "== DO POST CREATE =="
  ::fwk::doPostCreate ::fwk::src
  ::fwk::doPostCreate ::fwk::lib

  ::fwk::printDivider
  ::fwk::printCBM "== SAVE AND CLOSE =="
  ::fwk::tool::saveProject
  if {! $::fwk::GuiMode} {    # close project only in non GUI mode
    ::fwk::tool::closeProject
  }

  ::fwk::printDivider
  ::fwk::printSuccess "Project creation completed successfully\n"
}

# ==============================================================================
proc ::fwk::buildProject {} {

  puts ""
  ::fwk::printInfo "Running build of ( $::fwk::PrjBuildName ) project ...\n"

  # change directory to the current project build
  cd $::fwk::PrjBuildPath

  # open project
  ::fwk::tool::openProject $::fwk::PrjBuildName

  # create output folder
  ::fwk::createOutDir

  ::fwk::printCBM "== RE-GEN ADDRESS SPACE =="
  if { $::fwk::addr::TypesToGen != {} } {
    ::fwk::addr::main $::fwk::addr::TypesToGen
  } else {
    ::fwk::printInfo "Nothing to generate set in ::fwk::addr::TypesToGen"
    }

  puts ""
  ::fwk::printCBM "== DO ON BUILD =="
  # execute doOnBuild
  ::fwk::doOnBuild ::fwk::src

  puts ""
  ::fwk::printCBM "== BUILD =="
  # build project and close it afterwards
  ::fwk::tool::buildProject

  puts ""
  ::fwk::printCBM "== EXPORT ARTIFACTS =="
  # copy artifacts in the out directory
  ::fwk::tool::exportOut

  if {! $::fwk::GuiMode || ! $::fwk::ShellMode } {    # close project only in non GUI mode
    ::fwk::tool::closeProject
  }

  puts ""
  ::fwk::printCBM "== DO POST BUILD =="
  # execute doPostBuild
  ::fwk::doPostBuild

  ::fwk::printDivider
  ::fwk::printSuccess "Build completed successfully.\n"

}

proc ::fwk::postBuildProject {} {

  puts ""
  ::fwk::printInfo "Running postbuild of ( $::fwk::PrjBuildName ) project ...\n"

  # change directory to the current project build
  cd $::fwk::PrjBuildPath

  puts ""
  ::fwk::printCBM "== DO POST BUILD =="
  # execute doPostBuild
  ::fwk::doPostBuild

  puts "-------------------------------------------------\n"
  ::fwk::printSuccess "Postbuild completed (successfully).\n"

}

# ==============================================================================
proc ::fwk::simulateProject {} {

  puts ""
  ::fwk::printInfo "Running Simulation of ( $::fwk::PrjBuildName ) project ...\n"

  # change directory to the current project build
  cd $::fwk::PrjBuildPath

  # open project
  ::fwk::tool::openProject $::fwk::PrjBuildName

  ::fwk::printCBM "== SET SIM =="
  ::fwk::setSim

  ::fwk::printCBM "== SIMULATE =="
  ::fwk::tool::simProject

  # # execute doOnBuild
  # ::fwk::doOnSim
  # puts "-------------------------------------------------"

  ::fwk::tool::closeProject

  ::fwk::printDivider
  ::fwk::printSuccess "Simulation Run completed successfully.\n"
}

proc ::fwk::synthProject {} {

  puts ""
  ::fwk::printInfo "Running Synthesis of ( $::fwk::PrjBuildName ) project ...\n"

  ::fwk::printCBM "== SYNTHESIS =="
  ::fwk::tool::synthProject

  ::fwk::printDivider
  ::fwk::printSuccess "Synthesis completed successfully.\n"
}

# Clean project directory
# ------------------------------------------------------------------------------
proc ::fwk::cleanProject {PrjBuildName} {
  ::fwk::cleanPrj
  ::fwk::cleanOut
}

# ==============================================================================
# Removes $::fwk::PrjBuildPath files
# ------------------------------------------------------------------------------
proc ::fwk::cleanPrj {} {
  ::fwk::utils::cleanDir $::fwk::PrjBuildPath
}

# ==============================================================================
# Removes $::fwk::$::fwk::ProjectPath/out/${::fwk::PrjBuildName} files
# ------------------------------------------------------------------------------
proc ::fwk::cleanOut {} {
  ::fwk::utils::cleanDir $::fwk::ProjectPath/out/${::fwk::PrjBuildName}
}

# ==============================================================================
# Creates $::fwk::PrjBuildPath
# ------------------------------------------------------------------------------
proc ::fwk::createPrjDir {} {
  file mkdir $::fwk::PrjBuildPath
}

# ==============================================================================
# Creates $::fwk::$::fwk::ProjectPath/out/${::fwk::PrjBuildName}
# ------------------------------------------------------------------------------
proc ::fwk::createOutDir {} {
  file mkdir $::fwk::ProjectPath/out/${::fwk::PrjBuildName}
}

# ==============================================================================
# TODO: implement test
proc ::fwk::testProject {} {}

# ==============================================================================
# TODO: implement gui
proc ::fwk::guiProject {} {
  puts ""
  ::fwk::printInfo "Running GUI ( $::fwk::PrjBuildName ) project ...\n"
  set ::fwk::GuiMode 1

  # change directory to the current project build
  cd $::fwk::PrjBuildPath

  if { [info procs ::fwk::tool::startProjectGui] > 0 } {
    ::fwk::tool::startProjectGui $::fwk::PrjBuildName
  }

}

proc ::fwk::openSdk {} {
  puts ""
  ::fwk::printInfo "Opening SDK ( $::fwk::PrjBuildName ) ...\n"

  # change directory to the current project build
  cd $::fwk::PrjBuildPath

  if { [info procs ::fwk::tool::openSdk] > 0 } {
    ::fwk::tool::openSdk
  }

}

# ==============================================================================
# executes the code block $script in all defined module namespaces.
# recursively from last children to parent. The root namespace is $namespace.
# ------------------------------------------------------------------------------
proc ::fwk::iterateModules {ins script} {

  # foreach ins [namespace children $ns] {
  #   ::fwk::iterateModules $ins $script
  # }
  for {set i [expr {[llength $::fwk::SrcNamespacesList]-1}]} {$i>=0} {incr i -1} {
    set ns [lindex $::fwk::SrcNamespacesList $i]
    if { [string match ${ins}* $ns] } {

      if { [info exists ${ns}::IsSrcModule] } {
        set curDir [pwd]
        if { [info exists ${ns}::TclPath] } {
          cd [subst $${ns}::TclPath]
        }
        namespace eval $ns $script

        cd $curDir
      }
    }; # filter namespace list with input namespace
  }
}


proc ::fwk::iterateModulesProc { ins procedureName } {

  for {set i [expr {[llength $::fwk::SrcNamespacesList]-1}]} {$i>=0} {incr i -1} {
    set ns [lindex $::fwk::SrcNamespacesList $i]
    if { [string match ${ins}* $ns] } {

      if { [info exists ${ns}::IsSrcModule] } {
        set curDir [pwd]
        if { [info exists ${ns}::TclPath] } {
          cd [subst $${ns}::TclPath]
        }
        if { [info procs ${ns}::${procedureName}] > 0 } {
          ::fwk::printInfo "executing ${ns}::${procedureName}"
          set addmsg "{in [subst $${ns}::TclFile]}"
          set script "::fwk::catchHdl ${ns}::${procedureName} ${addmsg}"
          namespace eval $ns $script
        }
        cd $curDir
      }; # filter namespace list with input namespace
    }
  }
}


# ==============================================================================
## @brief executes doCreate procedure from all namespacess
# recursively from last children to parent, run on project create
# ------------------------------------------------------------------------------
proc ::fwk::doOnCreate {{ns ::fwk::src}} {
  ::fwk::iterateModulesProc $ns "doOnCreate"
}

proc ::fwk::doPostCreate {{ns ::fwk::src}} {
  ::fwk::iterateModulesProc $ns "doPostCreate"
}

# ==============================================================================
## @brief executes doOnBuild procedure from all namespacess
# recursively from last children to parent before tool build starts
# ------------------------------------------------------------------------------
proc ::fwk::doOnBuild {{ns ::fwk::src}} {
  ::fwk::iterateModulesProc $ns "doOnBuild"
}
# ==============================================================================
## @brief executes doPostBuild procedure from all namespacess
# recursively from last children to parent after build
# ------------------------------------------------------------------------------
proc ::fwk::doPostBuild {{ns ::fwk::src}} {
  ::fwk::iterateModulesProc $ns "doPostBuild"
}

# ==============================================================================
## @brief executes doOnSim procedure from all namespacess
# recursively from last children to parent before tool build starts
# ------------------------------------------------------------------------------
proc ::fwk::doOnSim {{ns ::fwk::src}} {
  ::fwk::iterateModulesProc $ns "doOnSim"
}
proc ::fwk::doPostSim {{ns ::fwk::src}} {
  ::fwk::iterateModulesProc $ns "doPostSim"
}
# ==============================================================================
## @brief executes setSim procedure from all namespacess
# recursively from last children to parent before tool build starts
# ------------------------------------------------------------------------------
proc ::fwk::setSim {{ns ::fwk::src}} {
  ::fwk::iterateModulesProc $ns "setSim"
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addSource {ns srcDict} {
  ::fwk::printDebug "${ns}::addSource \{${srcDict}\}"

  # --------------------------------------------------
  # check keys
  set sourceKeysList \
    {\
       File \
       Path \
       Type \
       Library \
       Template \
       Out \
       UsedIn \
       ProcOrder \
       Fileset \
       Virt \
       ToolArgs
    }
  set keys [dict keys $srcDict]
  foreach key $keys {
    if {[lsearch -exact ${sourceKeysList} $key] <0 } {
      ::fwk::printError "In $ns not valid key: $key\n
      addSource \{$srcDict\}\n"
      puts "Valid keys: ${sourceKeysList}"
      ::fwk::exit -1
    }
  }

  # TODO : add source Types veryfication

  # check if virtual file - do not run glob
  # e.g. does not exist at the addSource time
  set virtual 0
  if { [dict exists $srcDict Virt]  &&  [dict get $srcDict Virt]} { set vurtual 1 }
  # --------------------------------------------------
  # Check File and Path Key and resolve glob expresion
  if { [dict exists $srcDict File ] } { ; # File key
    if { $virtual == 0 } {
      if {[catch { set curFiles [glob -types f [dict get $srcDict File]] } resulttext ]} {
        ::fwk::printError "In $ns source file does not exist:\n
      addSource \{$srcDict\} \n
      Note: add `Virt True` key to bypass glob check and path normalization.\n"
        ::fwk::exit -1
      }
    }
  } elseif { [dict exists $srcDict Path ] } { ; # Path key
    if { $virtual == 0 } {
      if {[catch { set curFiles [glob -types d [dict get $srcDict Path]] } resulttext ]} {
        ::fwk::printError "In $ns source file does not exist:\n
      addSource \{$srcDict\} \n
      Note: add `Virt True` key to bypass glob check and path normalization.\n"
      ::fwk::exit -1
      }
    }
  } else { ; # error
    ::fwk::printError "No 'File' or 'Path' key in \n\{${srcDict}\}"
    ::fwk::exit -2
  }

  # --------------------------------------------------
  # Library: if not defined, default library is the ns name same as InstnaceName
  if { ![dict exists $srcDict Library] } {
    dict set srcDict Library [subst $${ns}::InstanceName]
  }

  # --------------------------------------------------
  # Template: if template key exist change path and add do template list
  set template 0
  if { [dict exists $srcDict Template] } {
    set template 1
  }

  # --------------------------------------------------
  # for each element on the glob list copy other dict keys
  # and add to Sources list or Template list
  foreach src $curFiles {
    set srcPath [file normalize $src]
    set tmpDict [dict replace $srcDict File $srcPath]
    if { $template } {

      if { [dict exists $srcDict Out] } {
        set tplOutFilePath [file normalize [dict exists $srcDict Out]]
      } else {
        set tplOutFile "[string range [file normalize $srcPath] [string length $::fwk::ProjectPath] end]"
        set tplOutFilePath ${::fwk::PrjBuildPath}${tplOutFile}
      }

      dict set tmpDict Out $tplOutFilePath
      lappend ${ns}::Templates $tmpDict

      dict remove $tmpDict Template
      dict remove $tmpDict Out
      dict set tmpDict File $tplOutFilePath
      lappend ${ns}::Sources $tmpDict

    } else { # no tempalte
      lappend ${ns}::Sources $tmpDict

    }
  }

}
# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addSources {ns} {
  ::fwk::printInfo "Adding ${ns}::Sources"
  if { [llength [subst $${ns}::Sources]] == 0} {
    ::fwk::printWarning "In ${ns} Sources list empty"
    return
  }
  ::fwk::tool::addSources [subst $${ns}::Sources]
}

proc ::fwk::addSourcesRec {{ns ::fwk::src}} {
  ::fwk::iterateModules $ns "addSources"
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::procTemplates {ns} {
  ::fwk::printInfo "Processing Templates ${ns}::Templates"

  set tplBackendList {"Jinja2"}

  foreach templateDict [subst $${ns}::Templates] {
    set srcFile [dict get $templateDict File]
    set outFile [dict get $templateDict Out]

    file mkdir [file dirname $outFile]
    switch [string tolower [dict get $templateDict Template]] {
      "jinja2" {
        # create json type context
        set context {}
        foreach {key value} [concat [array get ${ns}::Config] [array get ${ns}::Info]] {
          if {[llength $value] == 1} {
            lappend context "\"$key\": \"$value\""
          } else {
            set tmpList [list "\"$key\": " "\[\"" [join $value "\", \""] "\"\]" ]
            lappend context [join $tmpList ""]
          }
        }
        set jinjaArgs [list [join $context ", "]]
        # unset python environment var temporary to be able to run in tools like Vivado
        # Vivado has bundled python which might be in different version then venv
        # pick shell python to run python scripts
        array set orig_env [array get ::env "PYTHON*"]
        array unset ::env "PYTHON*"
        set PYTHON [exec bash -c "which python"]

        if { [catch {eval exec >&@stdout \
                       [list $PYTHON ${::fwk::Path}/fwk_jinja2.py $jinjaArgs $srcFile $outFile] \
                     } resulttext] } {
          ::fwk::printError "Jinja2 template engine failed:"
          puts "jinja2 $jinjaArgs $srcFile\n"
          set einfo $::errorInfo
          set ecode $::errorCode
          ::fwk::printError "\n $einfo"
          ::fwk::exit -2
        } else {
          ::fwk::printInfo "Rendered template to $outFile"
        }
        array set ::env [array get orig_env]
      }

      default {
        ::fwk::printError "Not supported Template backend. Available options: ${tplBackendList} "
      }
    }
  }
}

proc ::fwk::procTemplatesRec {{ns ::fwk::src}} {
  ::fwk::iterateModules $ns "procTemplates"
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addFwkIp {InstanceName} {

  ::fwk::printInfo "executing ::fwk::addFwkIp in $::fwk::ToolType with ::fwk::ip::${InstanceName}"

  set ipNs ::fwk::ip::${InstanceName}
  set IpName  [subst $${ipNs}::ModuleName]
  set IpVendor  [subst $${ipNs}::Vendor]
  set IpLibrary "fwk"
  set InsName $IpName
  set IpConfig {}

  ::fwk::tool::addIp $IpVendor $IpLibrary $IpName $InsName $IpConfig

}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addGenIPSources {{ns ::} args} {

  set sources {}

  # go to namespace tcl file location (sources have relative path to tcl)
  set curDir [pwd]

  if { [::fwk::checkSources $ns args sources] } { return }

  ::fwk::tool::addGenIPs $args $sources

  cd $curDir

}

# ==============================================================================
## print project info
# ------------------------------------------------------------------------------
proc ::fwk::printProjectInfo {} {

  ::fwk::utils::cprint "-------------------------------------" {".*" -fg cyan -style bright}
  ::fwk::utils::cprint "Loaded Configuration:"                 {".*" -fg cyan -style none}
  ::fwk::utils::cprint "-------------------------------------" {".*" -fg cyan -style bright}
  ::fwk::utils::cprint "ProjectName : $::fwk::ProjectName "  {"ProjectName" -fg white -style none}
  ::fwk::utils::cprint "ProjectConf : $::fwk::ProjectConf "  {"ProjectConf" -fg white -style none}
  ::fwk::utils::cprint "ProjectTcl  : $::fwk::ProjectTcl "   {"ProjectTcl"  -fg white -style none}
  ::fwk::utils::cprint "ToolType    : $::fwk::ToolType"      {"ToolType"    -fg white -style none}

  if { [info exists ::env(FWK_HW_FILE)] } {
    ::fwk::utils::cprint "HwFile      : $::fwk::HwFile"        {"HwFile" -fg white -style none}
  }

  ::fwk::utils::cprint "....................................." {".*" -fg white -style none}
  ::fwk::utils::cprint "Project Version : $::fwk::VerString" {".*" -fg white -style bright }
  ::fwk::utils::cprint "-------------------------------------" {".*" -fg cyan -style bright}

}

# ==============================================================================
## returns constant variables in indexed array
# ------------------------------------------------------------------------------
proc ::fwk::parseVhdlConfigFile {ns varName vhdlFile} {

  set name {}
  set type {}
  set value {}

  ::fwk::printInfo "parsing vhdl configuration file $vhdlFile into ${ns}::${varName}"

  set curDir [pwd]
  cd [file dirname [file normalize [subst $${ns}::TclFile]]]
  set vhdlFile [file normalize "$vhdlFile"]
  cd $curDir

  if {[ file isfile $vhdlFile ]} {
    set Cfg [open $vhdlFile r]
    while { [gets $Cfg line] >= 0 } {
      set match [regexp -expanded {(^\s*\t*--.*)} $line]
      if {1 != $match} {
        set match [regexp {constant\s+(\w+).*:.*(std_logic_vector).*:=\s*([Xx]*)"([0-9a-fA-F]+)"\s*;(.*)} $line totalmatch name type hex value]
        if {1 == $match} {
          if { $hex == "x" || $hex == "X" } {
            #set Constants($name) [ expr "0x$value" ]
            set Constants($name) 0x$value
          } else {
            binary scan [binary format B64 [format %064s $value]] W dec
            set Constants($name) $dec
          }
        }
        set match [regexp {constant\s+(\w+).*:.*(integer|natural|real).*:=\s*([_\d\(\)\+\-\*\/\.]+)\s*;(.*)} $line totalmatch name type value]
        if {1 == $match} {
          regsub -all "_" $value "" value
          set Constants($name) [ expr $value ]
        }
        set match [regexp {constant\s+(\w+).*:.*(string).*:=.*\"(.*)\".*} $line totalmatch name type value]
        if {1 == $match} {
          set Constants($name) $value
        }
      } else {
        #puts "Line is commented: $line";
      }
    }

    array set ${ns}::${varName} [array get Constants]
    return [array get Constants]
    close $Cfg
  } else { ::fwk::printError "::fwk::parseVhdlConfigFile cannot open $vhdlFile" ; ::fwk::exit -7 }
  return
}


# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::listSubmodules {} {

  ::fwk::utils::cprint "         Git Submodules" {".*" -fg cyan -style bright}
  ::fwk::utils::cprint "-------------------------------------" {".*" -fg cyan -style bright}
  puts [format "%-*s %-*s %-*s" 30 "Module path" 24 "Version" 24 "on branch"]
  ::fwk::printDivider
  if {[catch {set gitInfo [exec git -C $::fwk::ProjectPath submodule status]} msg]} { return }
  foreach {sha modulePath version} $gitInfo {
    if { ![file exists "$::fwk::ProjectPath/$modulePath/.git"] } {
      puts [format "%-*s %-*s %-*s" 30 ${modulePath} 24 "Not Initialized submodule"  24 ""]
      puts "..................................................................."
      ::fwk::printError "Not intilized submodule ${modulePath} run 'git submodule update --init ${modulePath}'"
      ::fwk::exit -1
    }
    # puts "$module $version"
    set branch [exec git -C $::fwk::ProjectPath/$modulePath branch --show-current]
    # if { $branch == "" } { set branch "HEAD detached" }
    set version [::fwk::getSubmoduleVersion $modulePath]
    puts [format "%-*s %-*s %-*s" 30 ${modulePath} 24 ${version} 24 ($branch)]
  }
  # include external libs info
  if {[info exists ::fwk::OsvvmPath] } {
    if { [string first $::fwk::ProjectPath $::fwk::OsvvmPath ]} { # include only if not within the project path
      puts "Include: $::fwk::OsvvmPath : "
      if {[catch {set branch [exec git -C $::fwk::OsvvmPath branch --show-current]} msg]} { return }
      if {[catch {set gitinfo [exec git -C $::fwk::OsvvmPath describe --long --tags --first-parent]} msg]} { return }
      puts [format "%-*s %-*s %-*s" 30 "osvvm" 24 ${gitinfo} 24 ($branch)]
    }
  }
  ::fwk::printDivider
#  [format "%-*s %*s %*s %*s %*s %*s %*s %*s %*s" 60 ${putline} 10 ${size} 12 ${address} 12 ${realsize} 6 ${bar} 6 ${width} 6 ${fracbits} 4 ${signed} 4 ${rwAccess}]
}


proc ::fwk::getSubmoduleVersion {modulePath} {

  set rc [catch {set gitinfo [exec git -C $::fwk::ProjectPath/$modulePath describe --tags --first-parent --dirty]} msg]
  if { $rc != 0 } {
    # if no tags use commit count for HEAD
    set VerMajor 0
    set VerMinor 0
    set VerPatch 0
    set rc [catch {set gitinfo [exec git -C $::fwk::ProjectPath/$modulePath rev-list HEAD --count]} msg]
    if { $rc == 0 } {
      set VerCommits $msg
      set version "0.0.0-$msg"
    } else {
      set VerCommits 0
    }
    set rc [catch {set gitinfo [exec git -C $::fwk::ProjectPath/$modulePath rev-parse --short HEAD]} msg]
    if { $rc == 0 } {
      set VerShasum [format %08x 0x$msg]
    } else {
      set VerShasum [format %08x 0]
      ::fwk::printWarning "no Git repository nor commits available to generate version\n"
    }
    set version "$VerMajor.$VerMinor.$VerPatch-$VerCommits-g$VerShasum"
  } else {
    #regexp {.*(\d+)\.(\d+)\.(\d+)[-]*(\d*).*} $msg match VerMajor VerMinor VerPatch VerCommits
    set version $msg
  }

  return $version
}
# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::getVersion {{ns ::fwk}} {

  set now [clock seconds]
  set ${ns}::Timestamp $now

  # go to namespace tcl file location (sources have relative path to tcl)
  set curDir [pwd]
  if {[llength [namespace which -variable ${ns}::ProjectPath]]} {
    # fwk namespace is special, we want the version of the project here,
    # not the fwk version. It is also the only namespace with a ProjectPath
    # variable.
    cd [subst $${ns}::ProjectPath]
  } elseif {[llength [namespace which -variable ${ns}::Path]]} {
    # regular source modules have the Path variable
    cd [subst $${ns}::Path]
  } else {
    # if no Tcl in namespacess then main namespaces
    # give version of superproject
    cd $::fwk::ProjectPath
  }

  set VerMajor 0
  set VerMinor 0
  set VerPatch 0
  set VerCommits 0
  set Semantic 1
  set rc [catch {set gitinfo [exec git describe --long --tags --first-parent]} msg]
  if { $rc != 0 } {
    # if no tags use commit count for HEAD
    set rc [catch {set gitinfo [exec git rev-list HEAD --count]} msg]
    if { $rc == 0 } {
      set VerCommits $msg
    } else {set VerCommits 0}
  } else {
    set match 0
    regexp {(\d+)\.(\d*)[.]*(\d*)[a-zA-Z0-9_]*[-]+(\d*)[-]+g[0-9a-fA-F]+} $msg match VerMajor VerMinor VerPatch VerCommits
    if { $match == 0 } {
      ::fwk::printWarning "NO VERSION. Latest TAG has no minimal semantic versioning <Major.Minor> syntax. e.g. 1.2.0, Only version string set to: $msg";
      set NoSemVer $msg;
      set Semantic 0}
    if { $VerPatch == "" } { set VerPatch 0;  ::fwk::printWarning "Version Tag has no Patch Level. Setting Patch = 0"}
  }

  set rc [catch {set gitinfo [exec git rev-parse --short HEAD]} msg]
  if { $rc == 0 } {
    set VerShasum [format %08x 0x$msg]
  } else {
    set VerShasum [format %08x 0]
    ::fwk::printWarning "no Git repository nor commits available to generate version\n"
  }
  set ::fwk::Hostname [info hostname]
  set ::fwk::CI 0
  if { [info exists ::env(CI)]} {
    if { $::env(CI) == true } {
      set ::fwk::CI 1
      ::fwk::printInfo "Running in CI at $::fwk::Hostname"
    }
  }
  # change shasum to inform about CI and not main/master branch
  if { $::fwk::CI == 0} {
    set VerShasum [expr 0x${VerShasum} + 0x20000000]
    set VerShasum [format %08x $VerShasum]
  }
  set ::fwk::Branch ""
  set rc [catch {set gitinfo [exec git branch --show-current ]} msg]
  if { $rc == 0 } {
    if { $msg != "main" && $msg != "master" } {
      set VerShasum [expr 0x${VerShasum} + 0x10000000]
      set VerShasum [format %08x $VerShasum]
      set ::fwk::Branch $msg
    }
  }
  cd $curDir

  # We use Semantic Versioning 2.0
  set ${ns}::VerMajor    [scan $VerMajor %d]
  set ${ns}::VerMinor    [scan $VerMinor %d]
  set ${ns}::VerPatch    [scan $VerPatch %d]
  set ${ns}::VerCommits  [scan $VerCommits %d]
  set ${ns}::VerShasum   "0x${VerShasum}"
  if { 1 == $Semantic } {
    set ${ns}::VerString "$VerMajor.$VerMinor.$VerPatch-$VerCommits-g$VerShasum"
  } else { set ${ns}::VerString $NoSemVer }
  set ${ns}::VerHex "[format %02x [subst $${ns}::VerMajor]][format %02x [subst $${ns}::VerMinor]][format %02x [subst $${ns}::VerPatch]][format %02x [subst $${ns}::VerCommits]]"
  set ${ns}::Ver [expr 0x[subst $${ns}::VerHex]]
  # add additiona information to version string
  if {$::fwk::Branch != "" } { # include branch name if not CI
    regsub -all {/} ${::fwk::Branch} {-} subbed
    set ${ns}::VerString "[subst $${ns}::VerString]-$subbed"
  }
  if {$::fwk::CI == 0 && ${ns} == "::fwk" } {     # include host name if not CI
    set ${ns}::VerString "[subst $${ns}::VerString]-${::fwk::Hostname}"
  }
  # puts " DEBUG: [subst $${ns}::VerString] : [subst $${ns}::VerHex]"
  if { $ns == "::fwk" } {
    set ::env(FWK_PRJ_VERSION) [format 0x%08x $::fwk::Ver]
    set ::env(FWK_PRJ_SHASUM)  [format 0x%08x $::fwk::VerShasum]
    set ::env(FWK_PRJ_VERSION_COMMITS) "$::fwk::VerCommits"
    set ::env(FWK_PRJ_VERSION_STRING) "$::fwk::VerString"
    set year  [scan [clock format $::fwk::Timestamp -format "%y"] %d]
    set month [scan [clock format $::fwk::Timestamp -format "%m"] %d]
    set day   [scan [clock format $::fwk::Timestamp -format "%d"] %d]
    set hour  [scan [clock format $::fwk::Timestamp -format "%H"] %d]
    set min   [scan [clock format $::fwk::Timestamp -format "%M"] %d]
    set sec   [scan [clock format $::fwk::Timestamp -format "%S"] %d]
    set ::fwk::TimestampXilinx [expr { \
                                         $sec + $min * int(pow(2, 6)) + \
                                         $hour * int(pow(2, 12)) + \
                                         $year * int(pow(2, 17)) + \
                                         $month * int(pow(2, 23)) + \
                                         $day * int(pow(2, 27)) \
                                      }] ;
  }
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::genPrjVerFile {{ns ::fwk} fileType fileName} {

  ::fwk::getVersion ::fwk

  switch $fileType {
    VHDL {

      set pkgName [file rootname [file tail $fileName]]

      set templateFile [open [file join $::fwk::ProjectPath fwk tpl tpl_pkg_prj_version.vhd] r]
      set template [read $templateFile]
      close $templateFile

      regsub -all {{PKG_NAME}} $template $pkgName subbed
      regsub -all {{PRJ_VERSION}} $subbed [format %08x $::fwk::Ver] subbed
      regsub -all {{PRJ_TIMESTAMP}} $subbed [format %08x $::fwk::Timestamp] subbed
      regsub -all {{PRJ_SHASUM}} $sub bed [format %08x $::fwk::VerShasum] subbed

      set curDir [pwd]
      cd [subst $${ns}::TclPath]

      set verFile [open $fileName w]
      ::fwk::printInfo "Created version file: $fileType $fileName"
      puts $verFile $subbed
      close $verFile
      cd $curDir
    }
    default {
      ::fwk::printError "::fwk::genVerFile not supported fileType: $fileType\n"
    }
  }

}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::genModVerFile {{ns} fileType fileName} {

  ::fwk::getVersion ${ns}

  switch $fileType {
    VHDL {

      set pkgName [file rootname [file tail $fileName]]

      set templateFile [open [file join $::fwk::ProjectPath fwk tpl tpl_pkg_version.vhd] r]
      set template [read $templateFile]
      close $templateFile

      regsub -all {{PKG_NAME}} $template $pkgName subbed
      regsub -all {{VERSION}} $subbed [format %08x [subst $${ns}::Ver]] subbed
      regsub -all {{TIMESTAMP}} $subbed [format %08x [subst $${ns}::Timestamp]] subbed

      set curDir [pwd]
      cd [subst $${ns}::TclPath]

      set verFile [open $fileName w]
      ::fwk::printInfo "Created version file: $fileType $fileName"
      puts $verFile $subbed
      close $verFile
      cd $curDir
    }
    default {
      ::fwk::printError "::fwk::genVerFile not supported fileType: $fileType\n"
    }
  }
}

# ==============================================================================
# ------------------------------------------------------------------------------
# add a not the the address space tree  in given namepsace, fixed tree variable AddressSpace
# checks keys
proc ::fwk::addAddressSpace {ns argDict} {

  # --------------------------------------------------
  # check keys
  set keysList \
    {\
       Type \
       Id \
       Name \
       File \
       Parent \
       Channel \
       Address \
       Size \
       Count \
       ChannelSel
    }
  set keys [dict keys $argDict]
  foreach key $keys {
    if {[lsearch -exact ${keysList} $key] <0 } {
      ::fwk::printError "In $ns not valid key: $key\n
      addAddresSpace \{$argDict\}\n"
      puts "Valid keys: ${keysList}"
      ::fwk::exit -1
    }
  }
  #key Id madnatory
  if { ![dict exists $argDict Id] } {
    ::fwk::printError "${ns}::addAddressSpace Key Id is mandatory.\n\
     Node: \{$argDict\}"
    ::fwk::exit -1
  }

  # normalize file path
  if { [dict exists $argDict File ]} { ; # File key
    if {[catch { set curFile [glob -types f [dict get $argDict File]] } resulttext ]} {
      ::fwk::printError "In $ns address space file does not exist:\n
      addAddresSpace \{$argDict\}"
      ::fwk::exit -1
    }
    dict set argDict File [file normalize $curFile]
  }
  if { ![dict exists $argDict Parent] } {
    dict set argDict Parent ""
  }
  if { ![ dict exists $argDict Channel] } {
    dict set argDict Channel 0
  }
  if { ![ dict exists $argDict ChannelSel] } {
    # dict set argDict ChannelSel [dict get $argDict Channel]
    set channel_sel [dict get $argDict Channel]
  } else {
    set channel_sel [dict get $argDict ChannelSel]
    dict unset argDict ChannelSel
  }

  if { ![dict exists $argDict Countj] } {
    dict set argDict Count 0
  }
  # If no default instance name given, Id becomes default instance name
  if { ![dict exists $argDict Name] } {
    dict set argDict Name [dict get $argDict Id]
  }

  if { ![dict exists $argDict Address] } {
    dict set argDict Address 0x00000000
  } else {
    dict set argDict Address [::fwk::addr::decodeAddress [dict get $argDict Address]]
  }

  if { ![dict exists $argDict Size] } {
    if {[dict get $argDict Address] != 0} {
      # set Range automatically based on address - binary shift operation
      set address [dict get $argDict Address]
      set size 0
      for {set idx 0} { $idx < 32 } { incr idx } {
        set bit [expr {($address>>$idx) & 1}]
        if { $bit == 0 } {
          set Range [expr {$size + int(pow(2,$idx))}]
        } else {break;}
      }
    } else {
      dict set argDict Size 0xFFFFFFFFFFFFFFFF
    }
  } else {
    dict set argDict Address [::fwk::addr::decodeAddress [dict get $argDict Address]]
  }
  # set node namespace/module
  dict set argDict Ns $ns

  # --------------------------------------------------------
  # Add node to the tree
  # ceck and init tree

  ::fwk::addr::tree::init_tree ${ns}::AddressSpace
  # namespace qualifiers 
  puts [namespace qualifiers ${ns}::AddressSpace]
  dict set ${ns}::AddressSpace Ns $ns
  # check parent key
  # if not set check channel and assign parent as channel or 0
  set parent_id [dict get $argDict Parent]

  if { $parent_id == "" } {
    set parent_id [::fwk::addr::tree::add_root_node ${ns}::AddressSpace $argDict]
    dict set argDict Parent $parent_id
  }


  # package require dicttool
  # puts [dict print [subst $${ns}::AddressSpace]]

  # if type is module - merge trees.
  # o not add module but copy all its childrens with the optional instance name change
  if {[dict get $argDict Type] == "module" } {
    set parent_indx [::fwk::addr::tree::find_node_by_id [subst $${ns}::AddressSpace] $parent_id [dict get $argDict Channel]]
    if { $parent_indx < 0} {
      ::fwk::printError "Cannot find Parent \"$parent_id\" for node \n \{$argDict\}"
      ::fwk::exit -1
    }
    set subtree_ns "[dict get $argDict Ns]::[dict get $argDict Id]"
    set subtree [subst $${subtree_ns}::AddressSpace]
    # puts "Add module in ${ns}::AddressSpace module "
    set subtree_root_indx [dict get $subtree Channels $channel_sel RootNode]
    set childrenNumber [llength [dict keys [dict get $subtree Nodes $subtree_root_indx Children]]]
    # copy each of child - change instance name if only one child - else use default
    foreach child_indx [dict keys [dict get $subtree Nodes $subtree_root_indx Children]] {
      if {${childrenNumber} == 1 } {
        dict set subtree Nodes $child_indx Name [dict get $argDict Name]
      }
      ::fwk::addr::tree::add_subtree ${ns}::AddressSpace $parent_indx $subtree $child_indx
    }
  } else {
    # if no module just add child
    set indx [::fwk::addr::tree::add_child ${ns}::AddressSpace $argDict]
  }

}



# ==============================================================================
# debug print function
proc ::fwk::printDebug { message } {
  if { ![info exists ::env(FWK_PRINT_DEBUG) ]} {return}
  global ColorTerminal
  if {$ColorTerminal == 0 } {
    set infoMessage "# DEBUG: $message"
  } else {
    # Use ANSI escape code for purple (bright magenta)
    set infoMessage "\033\[1;35m# DEBUG: \033\[0m $message"
  }
  puts $infoMessage
}
proc ::fwk::printInfo { message } {
  global ColorTerminal
  if {$ColorTerminal == 0 } {
    set infoMessage "# INFO: $message"
  } else {
    set infoMessage "\033\[1;38;5;8m# INFO: \033\[0m $message"
    #set infoMessage "\33\[1;37m# INFO: \33\[0m $message"
  }
  puts $infoMessage
  #::fwk::utils::cprint $infoMessage {"# INFO:"  -style bright}
}
proc ::fwk::printInfoMLine { message } {
  global ColorTerminal
  if {$ColorTerminal == 0 } {
    set infoMessage "# INFO: [lindex $message 0]"
  } else {
    set infoMessage "\033\[1;38;5;8m# INFO: \033\[0m [lindex $message 0]"
    #set infoMessage "\33\[1;37m# INFO: \33\[0m $message"
  }
  puts $infoMessage
  foreach msg [lrange $message 1 end] {
    puts $msg
  }
  #::fwk::utils::cprint $infoMessage {"# INFO:"  -style bright}
}

proc ::fwk::printSuccess { message } {
  global ColorTerminal
  set infoMessage "# INFO: $message"
  if {$ColorTerminal == 0 } {
    puts $infoMessage
  } else {
    # ::fwk::utils::cprint $infoMessage {"(?i)(# INFO:|success|successfull)" -fg green -style bright}
    puts "\033\[1;32m$infoMessage\033\[0m"
  }
}

proc ::fwk::printWarning { message } {
  global ColorTerminal
  if {$ColorTerminal == 0 } {
    set infoMessage "# WARNING: $message"
  } else {
    set infoMessage "\033\[1;38;5;11m# WARNING: \033\[0m $message"
  }
  puts $infoMessage
  # set infoMessage "# WARNING: $message"
  # ::fwk::utils::cprint $infoMessage {"# WARNING:" -fg yellow -style bright}
}

proc ::fwk::printError { message } {
  global ColorTerminal
  if {$ColorTerminal == 0 } {
    set infoMessage "# ERROR: $message"
  } else {
    set infoMessage "\033\[1;38;5;1m# ERROR: \033\[0m $message"
  }
  puts $infoMessage
  # set infoMessage "# ERROR: $message"
  # ::fwk::utils::cprint $infoMessage {"# ERROR:" -fg red -style bright}
}


proc ::fwk::printCBM {message} {
  global ColorTerminal
  if {$ColorTerminal == 0 } {
    set infoMessage $message
  } else {
    set infoMessage "\033\[1;38;5;6m${message}\033\[0m"
  }
  puts $infoMessage
  # ::fwk::printc {-fg cyan -style bright} $message
}

proc ::fwk::printc {argv message} {
  global ColorTerminal
  if {$ColorTerminal == 0 } {
    puts $message
  } else {
    lappend args ".*"
    foreach arg $argv {
      lappend args $arg
    }
    ::fwk::utils::cprint $message $args
  }
}

proc ::fwk::printDivider {} {
  ::fwk::utils::cprint -------------------------------------------------------------------------------- {".*" -fg cyan -style bright}
}
################################################################################

# This procedure copy a file in the release folder (out).
# It changes the name to the standard one ${::fwk::Appname} keeping the same extension
# An optional suffix can be given
proc ::fwk::releaseFile {path {suffix ""}} {

  if {$suffix != ""} {
    set suffix "_$suffix"
  }

  set ext [file extension $path]
  set releasePath [file join $::fwk::ReleasePath $::fwk::ReleaseName$suffix$ext]

  ::fwk::printInfo "Releasing $::fwk::ReleaseName$suffix$ext"
  file copy -force $path $releasePath
}

# This procedure scans all the namespace instances, fetches the 'Config' variable,
# Converts it to a JSON string,
# Runs the Jinja2 using a custom wrapper script (fwk_jinja2.py)
# Adds the generated file path into the namespace variable "Sources" or "Vhdl"
# This procedure runs only if namespace has a variable called "addConfigAsHdl"
# Users can visualize the JSON string as prints on CLI by adding a namespace variable
# called "printConfigJson"

# This procedure gets called by ::fwk namespace at the end of INIT stage and the beginning of
# createProject stage.
proc ::fwk::genHdlFromConfig {} {

  proc exportConfig {ns} {

    if { [info exist ${ns}::addConfigAsHdl] } { # Check if module wants to export Config as VHDL
        if { [array exists ${ns}::Config] } { # Check if Config variable exists
          set tailName [namespace tail ${ns}]
          ::fwk::printInfo "Found Config parameters for ${tailName} module"
          # Create hdl_config directory inside the PrjBuildPath (if it doesn't exists already)
          if {![file exists $::fwk::PrjBuildPath/${::fwk::PrjBuildName}.hdl_config]} {
            file mkdir $::fwk::PrjBuildPath/${::fwk::PrjBuildName}.hdl_config
          }

          # Choose the path of the generated code, template and the
          set hdlConfigPath $::fwk::PrjBuildPath/${::fwk::PrjBuildName}.hdl_config/pkg_${tailName}_config.vhd
          set templatePath  ${::fwk::ProjectPath}/fwk/tpl/tpl_pkg_module_config.vhd.j2

          # Construct a JSON data that contains Config parameters
          set jsonData "\{ \"Config\" : \{"
          foreach {key value} [array get ${ns}::Config] {
            if {[llength $value] > 1} { # List detected (assuming a list has more than one element)
                # Initialize JSON string for the list
                set listJson "\["
                foreach item $value {
                    append listJson "${item} ,"
                }
                # Remove trailing comma and close the list
                set listJson [string trimright $listJson ,]
                append listJson "\]"
                append jsonData "\"list_$key\": $listJson ,"
                # append the string "list_" in front of the Config parameter if its nested (2D)
                # Because its hard to resolve list types inside Jinja templates...
                # Template will then check if parameter has "list_" in front of the Config Parameter
                # If yes, it will create JSON structure as following:
                # list_C_2D_PARAMETER": [7 ,6 ,5 ,4 ,3 ,2 ,1 ,0 ]

              } else { # Simple key-value pair or single-element list
                  # Checking if user placed a string value inside a Config variable
                  # If yes, we attach "string_" prefix to the JSON data that feed into the Jinja engine
                  if {![string is integer -strict $value] && ![string is double -strict $value] && [string length $value] > 0} {
                    append jsonData "\"string_$key\": \"$value\" ,"
                  } else {
                    append jsonData "\"$key\": \"$value\" ,"
                  }
              }
          }
          set jsonData [string replace $jsonData end end "\}"]
          # Users can request to see the generated Config JSON to debug or inspect via printConfigJson variable
          if { [info exist ${ns}::printConfigJson] } {
            set formatted_output [::fwk::format_json $jsonData]
            ::fwk::printDebug $formatted_output
          }

          # Append timestamp information to the JSON data that is going to be consumed by jinja
          set currentTime [clock seconds]
          set currentDate [clock format $currentTime -format "%Y-%m-%d"]
          set currentTimeFormatted [clock format $currentTime -format "%H:%M:%S %Z"]
          append jsonData ", \"moduleName\" : \"${tailName}\", \"date\" : \"${currentDate}\" , \"time\" : \"${currentTimeFormatted}\" \}"

          # Unset the Python that comes together with the Tool
          array set orig_env [array get ::env "PYTHON*"]
          array unset ::env "PYTHON*"
          # Call the fwk_jinja2.py script to render the HDL template
          set PYTHON [exec bash -c "which python"]
          eval exec >&@stdout [list $PYTHON $::fwk::Path/fwk_jinja2.py $jsonData $templatePath $hdlConfigPath]
          # TODO: maybe catch the error and show it with ::fwk::printError

          # Append the generated VHDL file to the Sources variable of the particular namespace
          # Note how FWK adds the generated file to the default library of the tool by leaving the string empty
          lappend ${ns}::Vhdl [list $hdlConfigPath "VHDL" ""]
          lappend ${ns}::Sources [list $hdlConfigPath "VHDL" ""]
          ::fwk::printInfo "Generated a pkg_${tailName}_config.vhd file and added to the Project"
        }
    }
  }

  # Procedure to recursively discover each child node
  proc discoverChildExportConfig {ns} {
    exportConfig $ns; # Export Configs out of each child
    set children [namespace children $ns]
    foreach child $children {
        discoverChildExportConfig $child ; # Recursively list sub-namespaces
    }
  }

  # Start from the ::fwk::src namespace and go deeper
  discoverChildExportConfig ::fwk::src
}


# Takes a JSON string and formats it for better prints
# param: json_str -> JSON string that Jinja2 template engine consumes.
proc ::fwk::format_json {json_str} {
    set indent_level 0
    set formatted_json ""
    set in_string 0

    foreach char [split $json_str ""] {
        switch -exact -- $char {
          "\{" {
            set indentation [string repeat "\t" $indent_level]
            append formatted_json "\n${indentation}${char}\n"
            incr indent_level
          }
          "\}" {
            incr indent_level -1
            set indentation [string repeat "\t" $indent_level]
            append formatted_json "\n${indentation}${char}\n"
          }
          "\[" {
            append formatted_json $char
            incr indent_level
          }
          "\]" {
            incr indent_level -1
            set indentation [string repeat "\t" $indent_level]
            append formatted_json "\n${indentation}${char}\n"
          }
          "," {
            set indentation [string repeat "\t" $indent_level]
            append formatted_json "${char}\n${indentation}"
          }
          ":" {
            append formatted_json " $char "
          }
          "\"" {
            append formatted_json $char
            set in_string [expr {!$in_string}]
          }
          default {
            if {$in_string} {
              append formatted_json $char
            } else {
              append formatted_json [string trim $char]
            }
          }
        }
    }

    return $formatted_json
}

