## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2019-2021 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2019-12-22
# @author Lukasz Butkowski <lukasz.butkowski@desy.de>
# @author Michael Buechler <michael.buechler@desy.de>
# --------------------------------------------------------------------------- #
# @brief
# Part of DESY FPGA Firmware Framework (fwk)
# FPGA firmware framework script handling address generation
# --------------------------------------------------------------------------- #

source fwk_addr_tree.tcl
source fwk_addr_rdl.tcl

namespace eval ::fwk::addr {
  variable ArtifactsPath
  variable TypesToGen {vhdl map}; # by default generate all
  variable TypesToAdd {}
  variable UserLibraryDirs {}
  variable MapMergeThr 256
  variable MapMergeChs {0 1 2 3 4 5 13 14 15 16}; # Default
}

# namespace with addtess space tree proceudres

namespace eval ::fwk::addr {

  # ------------------------------------------------------------------------------
  # allows to invoko main address space generateion
  proc ::fwk::addr::generate { {types {} } } {
    if { $types == {} } {
      ::fwk::printInfo "::fwk::addr::generate No argument provided, using TypesToGen variable"
      ::fwk::printInfo "TypesToGen: $::fwk::addr::TypesToGen"
      ::fwk::addr::main $::fwk::addr::TypesToGen
    } else {
      ::fwk::printInfo "::fwk::addr::generate Generating: $types"
      ::fwk::addr::main $types
    }
  }

  # ------------------------------------------------------------------------------
  proc main {args} {
    # decode and check arguments
    set addrType      [lindex $args 0]
    set addressSpace  [lindex $args 1]
    set projectName   [lindex $args 2]

    if { $addressSpace == "" } {
      upvar 0 ::fwk::src::AddressSpace AddressSpace
    } else {
      upvar 0 $addressSpace AddressSpace
    }
    if {$projectName==""} {set projectName $::fwk::PrjBuildName}

    # -------------------------------------------------------
    ::fwk::printInfo "Iterate over address space trees and merge"

    # constructTree AddressSpace

    ::fwk::printDivider
    ::fwk::printInfo "Constructed address space tree under ::fwk::AddressSpace"

    ::fwk::addr::tree::print_tree AddressSpace
    # --------------------------------------------------------
    # Artifacts
    variable ArtifactsPath
    set ArtifactsPath [file join ${::fwk::PrjPath} ${projectName} ${projectName}.desyrdl]
    if {![file exists $ArtifactsPath]} { file mkdir $ArtifactsPath }

    # generate RDL files - address space database
    ::fwk::addr::rdl::genRdlFiles AddressSpace $ArtifactsPath

    ::fwk::printInfo "List of .rdl files in project:"
    dict for {accessChannel files} $::fwk::addr::RdlFiles {
      ::fwk::printInfo "Access Channel $accessChannel:"
      dict for {nodeId filename} $files {
        set filename [string replace $filename 0 [expr {[string length ${::fwk::ProjectPath}/]-1}]]
        ::fwk::printInfo "  $nodeId : $filename"
      }
    }

    variable ::fwk::src::DesyRdlSrc {}

    # generate output if an argument provided
    if { $addrType != "" } {
      set TypesToGen $addrType

      ::fwk::addr::genAddrSrc
    }

  return 0
  }

  # ------------------------------------------------------------------------------
  # namespace export main
  namespace export generate
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addr::genAddrSrc {} {
  variable ArtifactsPath
  variable TypesToGen
  variable MapMergeThr
  variable MapMergeChs

  ::fwk::addr::runDesyRdl $TypesToGen

  # process generated File by DesyRDL
  # DesyRDL generates ouput per 1 access channel - one root - merge channels if needed
  # select output file, next use in-out converter
  foreach type $TypesToGen {
    switch $type {
      map {
        ::fwk::printInfo "Mapp files output path: $::fwk::ProjectPath/out/${::fwk::PrjBuildName}"
        # ::fwk::addr::runDesyRdl map
        foreach maptype {mapp mapt} {
          set mapList {}
          set mergedFileName [file join $::fwk::ProjectPath/out/${::fwk::PrjBuildName} ${::fwk::PrjBuildName}_${::fwk::VerString}.$maptype]
          set mergedMapfile  [open $mergedFileName w]
          # Injecting the map file version in the map file itself
          puts $mergedMapfile "@MAPFILE_REVISION ${::fwk::VerString}"
          ::fwk::printInfo "Creating global $maptype file"
          # copy mapfiles per access channel
          foreach accessChannel [dict keys $::fwk::addr::RdlFiles] {
            set tmpMapPath [file join $ArtifactsPath map]
            set fileName [file join $::fwk::ProjectPath/out/${::fwk::PrjBuildName} ${::fwk::PrjBuildName}_${::fwk::VerString}_ch${accessChannel}.$maptype]
            ::fwk::printInfo "Creating $maptype file from perspective of AccessChannel:${accessChannel}"
            set mapfile [open $fileName w]
            puts $mapfile "@MAPFILE_REVISION ${::fwk::VerString}"
            set infile [file join $tmpMapPath ch${accessChannel}.$maptype]
            set inchannel [open $infile]
            while {[gets $inchannel line] >= 0} {
              if { $line != "" } {
                puts $mapfile $line
                if { [lsearch $MapMergeChs $accessChannel] >= 0 } { # add to merge map list if channel on the list
                  lappend mapList $line
                }
              }
            };
            close $inchannel
            close $mapfile
          }
          # ----------------------------
          # remove duplicates from common list with condition, remove from sorted list using indexes
          set newMapList {}
          set mapListSort [lsort $mapList]
          set idx 0
          while { $idx < [llength $mapListSort] } {
            set currR     [lindex $mapListSort $idx]
            set currRName [lindex $currR 0]
            set currRSize [lindex $currR 3]
            set currRCh   [lindex $currR 4]
            set nextIdx   [expr {$idx+1}]
            set nextR     [lindex $mapListSort $nextIdx]
            set nextRName [lindex $nextR 0]
            set nextRSize [lindex $nextR 3]
            set nextRCh   [lindex $nextR 4]

            if { $currRName == $nextRName } { # duplicate
              # higher channel has prio if size > 512 - arbitrary number - make configurable if needed
              if { ($currRCh < $nextRCh && $currRSize > $MapMergeThr) || ($currRCh > $nextRCh && $currRSize < $MapMergeThr) } {
                # remove current
                set mapListSort [lreplace $mapListSort $idx $idx]
              } else {
                # remove next
                set mapListSort [lreplace $mapListSort $nextIdx $nextIdx]
              }
            } else {            # no more duplicates
              incr idx
            }
          }
          set mapListMeta {}
          set mapListNoMeta {}
          foreach line $mapListSort {
            set match_meta [regexp {^@} $line]
            if {$match_meta} {
              lappend mapListMeta $line
            }
            set match_reg [regexp {^[A-Za-z\.0-9_]} $line]; # filters empty lines and comments
            if {$match_reg} {
              lappend mapListNoMeta $line
            }
          }
          # put meta into map file
          foreach line $mapListMeta {
            puts $mergedMapfile $line
          }
          # after duplicates removal sort back based on address, interupt registers on top
          set mapListSort [lsort -index 4 -integer  [lsort -index 2 -integer [lsort -index 8 $mapListNoMeta]]]
          foreach line $mapListSort {
            puts $mergedMapfile $line
          }
          close $mergedMapfile
        }
      }
      vhdl {
        # ::fwk::addr::runDesyRdl vhdl
      }
      h {
        # ::fwk::addr::runDesyRdl h
        # copy h per access channel
        foreach accessChannel [dict keys $::fwk::RdlFiles] {
          set tmpHPath [file join $ArtifactsPath h]
          set fileName [file join $::fwk::ProjectPath/out/${::fwk::PrjBuildName} ${::fwk::PrjBuildName}_${::fwk::VerString}_ch${accessChannel}.h]
          set hfile [open $fileName w]
          set infile [file join $tmpHPath ch${accessChannel}.h]
          set inchannel [open $infile]
          while {[gets $inchannel line] >= 0} {
            puts $hfile $line
            if { $line == "#define __desyrdl_ch${accessChannel}__H__" } {
              puts $hfile "#define DESYRDL_CH${accessChannel}_VERSION_STRING \"${::fwk::VerString}\""
              puts $hfile "#define DESYRDL_CH${accessChannel}_VERSION [format 0x%08x ${::fwk::Ver}]"
              puts $hfile "#define DESYRDL_CH${accessChannel}_SHASUM  [format 0x%08x ${::fwk::VerShasum}]"
            }
          }
          close $inchannel
          close $hfile
        }
      }
      adoc {
        # ::fwk::addr::runDesyRdl adoc
      }
      default {
        puts  "Generate only RDL files"
        # puts "# WARNING: Not supported option ::fwk::addr::main $args"
        # ::fwk::exit -7
      }
    }
  }
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addr::addAddrSrc {} {
  variable ArtifactsPath
  variable TypesToAdd
  variable NameSpace

  foreach accessChannel [dict keys $::fwk::addr::RdlFiles] {
    # if {[namespace exists ::fwk::src::desyrdlch${accessChannel}]} { namespace delete ::fwk::src::desyrdlch${accessChannel}}
    foreach typeToAdd $TypesToAdd {
      switch $typeToAdd {
        vhdl {
          ::fwk::printInfo "Adding address space $TypesToAdd files to project workspace"
          set libsDirs [glob -type d -nocomplain -directory [file join $ArtifactsPath "vhdl"] *]

          foreach libDir $libsDirs {
            set library [file tail $libDir]
            ::fwk::addSource ::fwk::addr [dict create File [file join $libDir *] Library $library Type "vhdl"]
          }
        }
        cocotb {
          ::fwk::printInfo "Adding cocotb address space modules to pypath"

          ::fwk::addSource ::fwk::addr [dict create Path [file join $ArtifactsPath cocotb/ ] \
                                          Library "cocotb" Type "PYTHONPATH" UsedIn "cosimulation"]
        }
        h {
          ::fwk::printInfo "Adding generated header files into the Project Workspace AccessChannel:${accessChannel}"
          # Set the source and destination directories
          set sourceDir $::fwk::ReleasePath
          set destDir [file join $::fwk::PrjPath ${::fwk::PrjBuildName} include]
          file mkdir $destDir
          # Get a list of all header files in the source directory
          set headerFiles [glob -nocomplain -directory $sourceDir *_ch${accessChannel}.h]
          # Iterate through the list of header files and copy each one to the destination directory
          # Here the name of each header files are changed to 'desyrdl_ch(x).h"
          foreach headerFile $headerFiles {
            file copy -force $headerFile [file join ${destDir} desyrdl_ch${accessChannel}.h]
          }
        }
      }
    }
  }
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addr::runDesyRdl {outType} {
  variable ArtifactsPath
  variable TypesToAdd
  variable UserLibraryDirs

  set outPath [file join $ArtifactsPath]
  file mkdir $outPath
  ::fwk::utils::cleanDir [file join $outPath $outType] ; # clean out type
  ::fwk::printDivider
  ::fwk::printInfo "Running DesyRdl to generate $outType files: "
  # variable ::fwk::src::DesyRdlSrc {}

  # run DesyRDL separately for each access channel
  dict for {accessChannel files} $::fwk::addr::RdlFiles {

    set RdlFilenamesPerCh {}
    foreach nodeId [lsort -integer -decreasing [dict keys $files]] {
      set rdlFile [dict get $files $nodeId]
      # append to the list only if file does not exist
      if { [lsearch $RdlFilenamesPerCh $rdlFile] == -1 } {
        lappend RdlFilenamesPerCh $rdlFile
      }
    }

    # The Tcl interpreter might set these environment variables, causing
    # problems when running our own Python program below. A specific example is
    # Vivado <2020.2 which bundles Python 2.7 while Vivado 2020.2 bundles
    # Python 3.8.
    array set orig_env [array get ::env "PYTHON*"]
    array unset ::env "PYTHON*"

    set cmdToRun "desyrdl -o $outPath -i $RdlFilenamesPerCh -f $outType tcl"
    if [llength ${::fwk::addr::UserLibraryDirs}] {
      set cmdToRun "$cmdToRun -l"
      foreach userLibDir ${::fwk::addr::UserLibraryDirs} {
        set cmdToRun "$cmdToRun $userLibDir"
      }
    }

    # puts [exec bash -c $cmdToRun]
    # -----------------------------------------------------------------------------
    if { [catch { eval exec >&@stdout "$cmdToRun" } resulttext] } {
      ::fwk::printError "DesyRdl failed:"
      ::fwk::exit -2
    }

    set fp [open $ArtifactsPath/desyrdl.log "w"]
    puts $fp $resulttext
    close $fp

    array set ::env [array get orig_env]
  }
  ::fwk::printDivider
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::addr::decodeAddress {address_data} {
    set match [regexp {([0-9]+)([GMk]+)} $address_data val num unit]
    if {$match == 1} {
      if { $unit == "k" } { set unitMul 1024 }
      if { $unit == "M" } { set unitMul 1048576 } ; # 1024*1024
      if { $unit == "G" } { set unitMul 1073741824 } ; # 1024*1024*1024
      set address_data [format "0x%08X" [expr {$num * $unitMul - 1 }] ]
    }
    return $address_data
}
