## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2025 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2025-04-30
# @author Lukasz Butkowski <lukasz.butkowski@desy.de>
# --------------------------------------------------------------------------- #
# @brief
# Part of DESY FPGA Firmware Framework (fwk)
# FPGA firmware framework script for SystemRdl files generation and handling
# --------------------------------------------------------------------------- #


namespace eval ::fwk::addr::rdl {

  # ------------------------------------------------------------------------------
  proc genRdlFiles {tree_ artifactsPath} {
    upvar 1 $tree_ tree

    ::fwk::printDivider
    ::fwk::printInfo "Generating Rdl files:"

    set rdlFilesPath [file join ${artifactsPath} rdl ]
    if {![file exists $rdlFilesPath]} { file mkdir $rdlFilesPath }
    ::fwk::utils::cleanDir $rdlFilesPath

    set ::fwk::addr::RdlFiles {}

    dict for {channel_id channel_data} [dict get $tree Channels] {
      # direcotry per channel
      set chRdlFilePath [file join ${rdlFilesPath} ${channel_id} ]
      file mkdir $chRdlFilePath

      set root_indx [dict get $channel_data RootNode]
      genRdlForNodes $tree $root_indx $chRdlFilePath
    }
  }

  # ------------------------------------------------------------
  proc genRdlForNodes {tree nodeIndx path} {

    set node         [dict get $tree Nodes $nodeIndx]
    set nodeInstName [dict get $node Name]
    set nodeId       [dict get $node Id]
    set nodeNs       [dict get $node Ns]
    set channel      [dict get $node Channel]
    set filePath     [file join ${path} ${nodeInstName} ]
    set fileName     [file join ${filePath} ${nodeId}]
    set fileNameRdl  ${fileName}.rdl
    set fileNameVh   ${fileName}.vh

    file mkdir $filePath

    # generate vh file based on config
    set cfgVariables [array get ${nodeNs}::Config]
    if {[llength $cfgVariables] > 0} {
      genVHFromConfig $cfgVariables $fileNameVh
    } else { # create dummy file if no configuration provided
      set tmpFile [open $fileNameVh w]
      close $tmpFile
    }

    # Rdl file based on the node type
    switch [string toupper [dict get $node Type]] {
      RDL {
        file copy -force [file normalize [dict get $node File]] $fileNameRdl
      }
      TOP {
        set rdlFileHandle [open $fileNameRdl w]
        genRdlForTopNode $tree $nodeIndx $rdlFileHandle
        close $rdlFileHandle
      }
      NODE {
        set rdlFileHandle [open $fileNameRdl w]
        genRdlForNode $tree $nodeIndx $rdlFileHandle
        close $rdlFileHandle
      }
      IPX {
        set rdlFileHandle [open $fileNameRdl w]
        genRdlFromIPX $tree $nodeIndx $rdlFileHandle
        close $rdlFileHandle
      }
      IBUS {
        set rdlFileHandle [open $fileNameRdl w]
        genRdlFromIIFile $tree $nodeIndx $rdlFileHandle $cfgVariables
        close $rdlFileHandle
      }
      INTERCONNECT {
        set rdlFileHandle [open $fileNameRdl w]
        genRdlForInterconnect $tree $nodeIndx $rdlFileHandle
        close $rdlFileHandle
        }
      default {
        ::fwk::printError "Node $nodeId not handled by DesyRDL"
        ::fwk::exit -2;
      }
    }

    # add file to the list of files
    dict set ::fwk::addr::RdlFiles $channel $nodeIndx $fileNameRdl
    # iterate recursive over childs
    dict for {child_indx child_id} [dict get $node Children] {
      genRdlForNodes $tree $child_indx $path
    }
  }

  # ------------------------------------------------------------
  proc genRdlForTopNode {tree nodeIndx rdlFile} {
    set moduleName [dict get $tree Nodes $nodeIndx Name]
    # count items under root parent channel,
    # use bridge only if > 1
    set childrenNumber [llength [dict keys [dict get $tree Nodes $nodeIndx Children]]]
    set accessChannel [dict get $tree Nodes $nodeIndx Channel]
    puts $rdlFile "\naddrmap $moduleName \{"
    puts $rdlFile "  desyrdl_generate_hdl = false ;"
    puts $rdlFile "  desyrdl_access_channel = $accessChannel ;"
    puts $rdlFile "  desyrdl_interface = \"AXI4L\" ;" ; # anytype to suppress warning, no HDL generated
    if {${childrenNumber} > 1 } {
      puts $rdlFile "  bridge;"
    }
    # iterte over children
    genRdlForChildren $tree $nodeIndx $rdlFile

    puts $rdlFile "\} ;"

  }

  proc genRdlForInterconnect {tree nodeIndx rdlFile} {
    set moduleName [dict get $tree Nodes $nodeIndx Name]
    # count items under root parent channel,
    # use bridge only if > 1
    set accessChannel [dict get $tree Nodes $nodeIndx Channel]
    puts $rdlFile "\naddrmap $moduleName \{"
    puts $rdlFile "  desyrdl_generate_hdl = false ;"
    puts $rdlFile "  desyrdl_access_channel = $accessChannel ;"
    puts $rdlFile "  desyrdl_interface = \"AXI4L\" ;" ; # anytype to suppress warning, no HDL generated

    genRdlForChildren $tree $nodeIndx $rdlFile

    puts $rdlFile "\} ;"
  }

  # ------------------------------------------------------------------------------
  proc genRdlForChildren {tree nodeIndx rdlFile} {
    foreach child_indx [dict keys [dict get $tree Nodes $nodeIndx Children]] {
      set typeName     [dict get $tree Nodes $child_indx Id]
      set instanceName [dict get $tree Nodes $child_indx Name]
      set address      [dict get $tree Nodes $child_indx Address]
      set size         [dict get $tree Nodes $child_indx Size]
      set count        [dict get $tree Nodes $child_indx Count]
      set stride       [expr {$size + 1}]
      set properties   "" ; # possbie to use config to change instance parameters
      if { $count == 0 } {
        puts $rdlFile "  $typeName $properties $instanceName @$address;"
        # puts $rdlFile "  external mem {} channel_high_address @$stride;"
      } else {
        puts $rdlFile "  $typeName $properties $instanceName  \[$count\] @$address += $stride;"
      }
    }
  }

  # ------------------------------------------------------------------------------
  proc genRdlForNode {tree nodeIndx rdlFile} {
    set moduleName [dict get $tree Nodes $nodeIndx Name]
    # count items under root parent channel,
    # use bridge only if > 1
    set childrenNumber [llength [dict keys [dict get $tree Nodes $nodeIndx Children]]]
    set accessChannel [dict get $tree Nodes $nodeIndx Channel]
    set size [dict get $tree Nodes $nodeIndx Size]
    set mementries [expr {($size + 1) / 4}]

    puts $rdlFile "\naddrmap $moduleName \{"
    puts $rdlFile "  desyrdl_generate_hdl = false ;"
    puts $rdlFile "  desyrdl_access_channel = $accessChannel ;"
    puts $rdlFile  "  external mem \{"
    puts $rdlFile  "    mementries        = $mementries;"
    puts $rdlFile  "    memwidth          = 32 ;"
    puts $rdlFile  "  \} $moduleName;"
    puts $rdlFile "\} ;"
  }

  # ------------------------------------------------------------------------------
  proc genRdlFromIIFile {tree nodeIndx rdlFile config} {
    set moduleName [dict get $tree Nodes $nodeIndx Name]
    # count items under root parent channel,
    # use bridge only if > 1
    set childrenNumber [llength [dict keys [dict get $tree Nodes $nodeIndx Children]]]
    set address        [dict get $tree Nodes $nodeIndx Address]
    set accessChannel  [dict get $tree Nodes $nodeIndx Channel]
    set addressFile    [dict get $tree Nodes $nodeIndx File]
    set size           [dict get $tree Nodes $nodeIndx Size]

    set regs [parseIIFile $address $addressFile $config $accessChannel]

    puts $rdlFile "addrmap $moduleName \{"
    puts $rdlFile "  desyrdl_generate_hdl = false ;"
    puts $rdlFile "  desyrdl_access_channel = $accessChannel ;"
    puts $rdlFile "  desyrdl_interface  = \"IBUS\" ;"

    # find nodes with this node as parent, instantiate them
    genRdlForChildren $tree $nodeIndx $rdlFile

    # iterate over register list and add to rdl file
    foreach reg  $regs {
      set type       [lindex $reg 0]
      set name       [lindex $reg 1]
      set size       [lindex $reg 2]
      set address    [lindex $reg 3]
      set realsize   [lindex $reg 4]
      set width      [lindex $reg 5]
      set fracbits   [lindex $reg 6]
      set signed     [lindex $reg 7]
      set swAccess   [lindex $reg 8]
      set hwAccess   [lindex $reg 9]

      # decode access mode
      if { $swAccess == "ro" } { set swAccess "r"}
      if { $swAccess == "wo" } { set swAccess "w"}
      if { $hwAccess == "ro" } { set hwAccess "r"}
      if { $hwAccess == "wo" } { set hwAccess "w"}

      if { ($hwAccess == "w" || $hwAccess == "rw") && ($swAccess == "w" || $swAccess == "rw")} {
        set we "true"
      } else { set we "false" }
      # decode data type to [u]<int|(fixed<N>)>
      # not covered by II files: char, float, double, half
      if { $fracbits > 0 } {
        set data_type "fixed$fracbits"
      } else {
        set data_type "int"
      }
      # pretend "u" if unsigned
      if { $signed  == 0 } { set data_type "u$data_type" }

      if { $type == "mem" } {
        puts $rdlFile  "  external mem \{"
        puts $rdlFile  "    mementries         = $size ;"
        puts $rdlFile  "    memwidth           = 32 ;"
        puts $rdlFile  "    sw                 = $swAccess ;"
        puts $rdlFile  "    desyrdl_data_type  = \"$data_type\" ;"
        puts $rdlFile  "    default desyrdl_data_type = \"$data_type\" ;"
        puts $rdlFile  "    default regwidth   = 32 ;"
        puts $rdlFile  "    default fieldwidth = $width ;"
        puts $rdlFile  "    reg \{"
        puts $rdlFile  "      desc = \"IBUS memory element\";"
        puts $rdlFile  "      field \{\} data ;"
        puts $rdlFile  "    \} DATATYPE\[$size\] ;"
        puts $rdlFile  "  \} $name @$address ;"
      } elseif { $type == "reg" } {
        puts $rdlFile  "  reg \{"
        puts $rdlFile  "    default sw        = $swAccess ;"
        puts $rdlFile  "    default hw        = $hwAccess ;"
        puts $rdlFile  "    default we        = $we ;"
        puts $rdlFile  "    desyrdl_data_type = \"$data_type\" ;"
        puts $rdlFile  "    field \{"
        puts $rdlFile  "    \} data\[$width\] ;"
        puts $rdlFile  "  \} $name\[$size\] @$address ;\n"
      } else {
        puts $rdlFile  " // external placed over tcl \"$name\[$size\] @$address \""
      }
    }

    puts $rdlFile "\} ;"

  }

  # ------------------------------------------------------------------------------
  # parses II  file, calculates items addresses and eval variables
  proc parseIIFile {BaseAddress AddressFile Config AccessChannel} {

    set regs {}

    # variables passed as a list key value, convert to array for easier search
    array set ConfVars $Config
    set addrFile [ open [file normalize $AddressFile] r]

    # pars line by line
    while { [gets $addrFile line] >= 0 } {
      # set proper bar based on variable name
      # set match [regexp {VIIItemDeclList.*Bar(\d+)} $line match bar]
      # address for DMA description is in bytes: set bytes 1
      set bytes 4

      # search for II items
      set match [regexp {^\s*VII_(AREA|EXTB|WORD|BITS)} $line]
      if {1 == $match} {
        set reg     [split $line ","]
        set type    [string trim [lindex $reg 0]]
        set name    [string trim [lindex $reg 1]]
        set width   [string trim [lindex $reg 2]]
        set size    [string trim [lindex $reg 3]]
        set address [string trim [lindex $reg 6]]
        set signed  [string trim [lindex $reg 7]]
        set swAccess  [string trim [lindex $reg 4]]
        set hwAccess  [string trim [lindex $reg 5]]
        if {1 == [string match "VII_UNSIGNED*" $signed]} {
          set signed 0
        } else {
          set signed 1
        }

        if {1 == [string match "VII_WACCESS*" $swAccess]} {
          set swAccess "rw"
        } elseif {1 == [string match "VII_WNOACCESS*" $swAccess]} {
          set swAccess "ro"
        } elseif {1 == [string match "VII_RNOACCESS*" $swAccess]} {
          set swAccess "wo"
        } else {
          set swAccess "rw"
        }

        if {1 == [string match "VII_REXTERNAL*" $hwAccess] &&  $swAccess == "ro" } {
          set hwAccess "wo"
        } elseif {1 == [string match "VII_RINTERNAL*" $hwAccess]} {
          set hwAccess "ro"
        } elseif {1 == [string match "VII_REXTERNAL*" $hwAccess]} {
          set hwAccess "rw"
        } else {
          set hwAccess "rw"
        }

        if { $type == "VII_WORD" } {
          set type "reg"
        } elseif { $type == "VII_EXTB" } {
          set type "ext"
        } else { set type "mem" }

        set fracmatch [regexp {[-]{4}fracbits([-]*[a-zA-Z0-9_\*/\+]+)} $line -> fracbits]
        if {0 == $fracmatch} {
          set fracbits 0
        } else {
          set fraclist [split $fracbits "*/+"]
          foreach fracel $fraclist {
            if {0 == [regexp {^[-]*\d+$} $fracel]} {
              if {![info exists ConfVars($fracel)]} { puts "\n#ERROR: Cannot find $fracbits fracbits: $fracel variable for $name. Check config provided." ; ::fwk::exit -2;}
              regsub $fracel $fracbits $ConfVars($fracel) fracbits
            }
          }
          set fracbits [expr ($fracbits)]
        }

        # set width
        set widthlist [split $width "*/+-"]
        foreach widthel $widthlist {
          if {0 == [regexp {^\d+$} $widthel]} {
            if {![info exists ConfVars($widthel)]} { puts "\n#ERROR: Cannot find width: $widthel variable for $name. Check config provided." ; ::fwk::exit -2;}
            regsub $widthel $width $ConfVars($widthel) width
          }
        }
        set width [expr ($width)]

        # recalculate address
        set addlist [split $address "*/+-"]
        foreach addel $addlist {
          if {0 == [regexp {^\d+$} $addel]} {
            if {![info exists ConfVars($addel)]} { puts "\n#ERROR: Cannot find address: $addel variable for $name. Check config provided." ; ::fwk::exit -2;}
            regsub $addel $address $ConfVars($addel) address
          }
        }
        set address [expr  ($address) * $bytes]

        # recalculate realsize
        set sizelist [split $size "*/+-"]
        foreach sizeel $sizelist {
          if {0 == [regexp {^\d+$} $sizeel]} {
            if {![info exists ConfVars($sizeel)]} { puts "\n#ERROR: Cannot find $widthel variable for $name. Check config provided." ; ::fwk::exit -2;}
            regsub $sizeel $size $ConfVars($sizeel) size
          }
        }
        set realsize [expr ($size)*$bytes]
        # workaround for DMA constants
        # set size [expr $size / $sizediv]
        if { $size == 0 } { set size 1}
        set spaces [expr {50-[string length $name]}]
        #set cnt $modulecount($prefix)
        set reg [concat ${type} ${name} ${size} ${address} ${realsize} ${width} ${fracbits} ${signed} ${swAccess} ${hwAccess}]
        lappend regs $reg
      }
    }
    close $addrFile

    return $regs
  }
  # ==============================================================================
  # ------------------------------------------------------------------------------
  proc genRdlFromIPX {tree nodeIndx rdlFile} {
    set moduleName [dict get $tree Nodes $nodeIndx Name]
    # count items under root parent channel,
    # use bridge only if > 1
    set childrenNumber [llength [dict keys [dict get $tree Nodes $nodeIndx Children]]]
    set address        [dict get $tree Nodes $nodeIndx Address]
    set accessChannel  [dict get $tree Nodes $nodeIndx Channel]
    set IPComponent    [dict get $tree Nodes $nodeIndx Id]
    set size           [dict get $tree Nodes $nodeIndx Size]
    set mementries     [expr {($size + 1) / 4}]

    puts $rdlFile "\naddrmap $moduleName \{"
    puts $rdlFile "  desyrdl_generate_hdl = false ;"
    puts $rdlFile "  desyrdl_access_channel = $accessChannel ;"
    puts $rdlFile "  desyrdl_interface = \"AXI4L\" ;"

    if {![namespace exists ::ipx] } {
      ::fwk::printWarning "Not in Vivado: cannot generate address space for $IPComponent. Placing generig MEM regin in $IPComponent."
      puts $rdlFile  "  external mem \{"
      puts $rdlFile  "    mementries        = $mementries ;"
      puts $rdlFile  "    memwidth          = 32 ;"
      puts $rdlFile  "  \} MEM ;"
      puts $rdlFile "\} ;"
      return -1
    }

    if { [file exists $IPComponent] } {
      puts "Found IP ${IPComponent} file, using as IPX file"
      ipx::open_ipxact_file $IPComponent
    } else {
      set componentXml [::fwk::utils::findFiles $::fwk::PrjBuildPath ${IPComponent}.xml]
      if { [llength $componentXml] > 1 } { # pick fist on list if found multiple xmls
        ::fwk::printWarning "Found multiple IP xml files. $componentXml. Using first from the list."
        set componentXml [lindex $componentXml 0]
      }
      if { [file exists $componentXml] } {
        ::fwk::printInfo "IP: found ${IPComponent}.xml, using as IPX file: ${componentXml}"
        ipx::open_ipxact_file $componentXml
      } else {
        ::fwk::printWarning "Cannot find IP xml file: $IPComponent. Placing generig MEM region in $IPComponent based on Tcl entry."
        puts $rdlFile  "  external mem \{"
        puts $rdlFile  "    mementries        = $mementries ;"
        puts $rdlFile  "    memwidth          = 32 ;"
        puts $rdlFile  "  \} MEM ;"
        puts $rdlFile "\} ;"
        return
      }
    }

    set regs [ipx::get_registers -of_object [ipx::get_address_blocks * -of_objects [lindex [ipx::get_memory_maps * -of_objects [ipx::current_core]] 0]]]
    set regsNum [llength $regs]
    if { $regsNum == 0 } {
      ::fwk::printWarning "IP $IPComponent has no registers. Placing generig MEM region in $IPComponent based on Tcl entry."
      puts $rdlFile  "  external mem \{"
      puts $rdlFile  "    mementries        = $mementries ;"
      puts $rdlFile  "    memwidth          = 32 ;"
      puts $rdlFile  "  \} MEM ;"
      puts $rdlFile "\} ;"
      return
    } else {
      ::fwk::printInfo "Found $regsNum registers in ${IPComponent}.xml"
    }

    foreach reg $regs {
      #report_property -all $reg
      set regName [get_property NAME $reg]
      set regDesc [get_property DESCRIPTION $reg]
      set regDispName [get_property DISPLAY_NAME $reg]
      set regAddr [get_property ADDRESS_OFFSET $reg]
      set regAcc  [get_property ACCESS $reg]
      set regSize [get_property SIZE $reg]
      set name [string toupper $regName]
      # replace all strange chars to _
      regsub -all {\s+|[-+,.;%$#@!()\[\]]+} $name "_" name

      switch $regAcc {
        "read-only" {
          set rwAccess r
        }
        "read-write" {
          set rwAccess rw
        }
        "write-only" {
          set rwAccess w
        }
        default {
          set rwAccess rw
        }
      }

      puts $rdlFile  "  reg \{"
      puts $rdlFile  "    name = \"$regDispName\" ;"
      puts $rdlFile  "    desc = \"$regDesc\" ;"
      puts $rdlFile  "    default sw = $rwAccess ;"
      puts $rdlFile  "    default hw = r ;"
      puts $rdlFile  "    field \{"
      puts $rdlFile  "    \} data\[$regSize\] ;"
      puts $rdlFile  "  \} $name @$regAddr ;\n"
    }
    puts $rdlFile "\} ;"
    #  ipx::open_ipxact_file [get_property XML_FILE_NAME [get_ipdefs $ipDef]]
    #  puts [ipx::get_registers -of_object [ipx::get_address_blocks * -of_objects [ipx::get_memory_maps * -of_objects [ipx::open_core $componentXml]]]]
    ipx::unload_core
    # pars line by line
  }


  proc genVHFromConfig {Config fileName} {
    # we want to align variables in the config file below, so first
    # obtain the maximum length of a config variable name
    set maxlen 0
    dict for {key value} $Config {
      set len [string length $key]
      if {$len > $maxlen} {
        set maxlen $len
      }
    }

    set fileHandle [open [file normalize $fileName] w]
    ::fwk::printInfo "Writing configuration constants to $fileName"
    dict for {key value} $Config {
      # - for left-justified
      # * to take the width from an arg (here: $maxlen)
      # s for no conversion, just a string
      set match [regexp {^D_.*} $key totalmatch]
      if {![string is double -strict $value] && $match==0} { set value "\"$value\""}; # put stings in double quote
      set line [format "`define %-*s %s" $maxlen $key $value]
      puts $fileHandle $line
    }
    close $fileHandle
  }
}
