## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2025 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2025-04-30
# @author Lukasz Butkowski <lukasz.butkowski@desy.de>
# --------------------------------------------------------------------------- #
# @brief
# Part of DESY FPGA Firmware Framework (fwk)
# FPGA firmware framework script handling address tree construction
# --------------------------------------------------------------------------- #

namespace eval ::fwk::addr::tree {

  # ------------------------------------------------------------
  proc init_tree {tree_} {
    upvar 1 $tree_ tree
    # intialize if tree empty,
    if {! [dict exists $tree NodeIndex]} {
      dict set tree NodeIndex -1
    }
    if {! [dict exists $tree Channels]} {
      dict set tree Channels {}
    }
    if {! [dict exists $tree Nodes]} {
      dict set tree Nodes {}
    }
  }

  # ------------------------------------------------------------
  proc generate_node_indx {tree_} {
    upvar 1 $tree_ tree
    set counter [dict get $tree NodeIndex]
    incr counter
    dict set tree NodeIndex $counter
    return "$counter"
  }

  # ------------------------------------------------------------
  proc find_node_by_name {tree name {channel 0}} {
    foreach node_indx [dict keys [dict get $tree Nodes]] {
      if { [dict get $tree Nodes $node_indx Channel] == $channel && \
           [dict get $tree Nodes $node_indx Name] eq $name} {
        return $node_indx
      }
    }
    return -1
  }

  # ------------------------------------------------------------
  proc find_node_by_id {tree id {channel 0}} {
    foreach node_indx [dict keys [dict get $tree Nodes]] {
      if { [dict get $tree Nodes $node_indx Channel] == $channel && \
           [dict get $tree Nodes $node_indx Id] eq $id } {
        return $node_indx
      }
    }
    return -1
  }

  # ------------------------------------------------------------
  proc add_node {tree_ node} {
    upvar 1 $tree_ tree
    set indx [generate_node_indx tree]
    set channel [dict get $node Channel]
    set parent_id [dict get $node Parent]
    if { $parent_id eq ""} {
      # check if root node exist for this channel
      # only one root allowed in channel
      if { [dict exists $tree Channels $channel]} {
        ::fwk::printError "Only one root node allowed per access channel.\n\
 Channel $channel has already a root node: [dict get $tree Channels $channel].\n\
 Please change Channel or Parent name for the node:\n \{$node\}\n"
        ::fwk::exit -1
      }
      dict set tree Channels $channel RootNode $indx
      set parent_indx ""
    } else {
      set parent_indx [find_node_by_id $tree $parent_id $channel]
      set channel [dict get $tree Nodes $parent_indx Channel]
    }
    dict set node Channel $channel
    dict set node ParentIndx $parent_indx
    dict set node Children {}
    dict set tree Nodes $indx $node
    return $indx
  }

  # ------------------------------------------------------------
  proc add_root_node {tree_ node} {
    upvar 1 $tree_ tree
    set channel [dict get $node Channel]
    set ns [dict get $node Ns]
    if {![dict exists $tree Channels $channel]} {
      dict set chnode Type "top"
      dict set chnode Id   "ch${channel}_top"
      dict set chnode Name "ch${channel}_top"
      dict set chnode Parent ""
      dict set chnode Channel $channel
      dict set chnode Address 0
      dict set chnode Size 0xFFFFFFFFFFFFFFFF
      dict set chnode Ns $ns
      add_node tree $chnode
    }
    return "ch${channel}_top"
  }

  # ------------------------------------------------------------
  proc add_child {tree_ node} {
    upvar 1 $tree_ tree
    set parent_id [dict get $node Parent]
    set channel [dict get $node Channel]
    set parent_indx [find_node_by_id $tree $parent_id $channel]
    # puts $tree
    if {![dict exists $tree Nodes $parent_indx]} {
      ::fwk::printError "Parent node $parent_id ($parent_indx) does not exist in the tree channel $channel.\nNode: \{$node\}"
      ::fwk::exit -1
    }
    set child_indx [add_node tree $node]
    dict set tree Nodes $parent_indx Children $child_indx [dict get $tree Nodes $child_indx Name]
    return $child_indx
  }

  # ------------------------------------------------------------
  proc add_subtree {tree_ parent_indx subtree root_indx} {
    upvar 1 $tree_ tree

    if {![dict exists $tree Nodes $parent_indx]} {
      ::fwk::printError "add_subtree: Parent node ($parent_indx) does not exist in the tree."
      ::fwk::exit -1
    }

    # ------------------------------------------------------------
    proc copy_node {tree_ parent_indx subtree_ node_indx} {
      upvar 1 $tree_ tree
      upvar 1 $subtree_ subtree

      set node [dict get $subtree Nodes $node_indx]
      if {[dict get $node Type] != "top" } {
        dict set node Parent [dict get $tree Nodes $parent_indx Id]
        dict set node Channel [dict get $tree Nodes $parent_indx Channel]
        set parent_indx [add_child tree $node]
        puts "ading node $node_indx : [dict get $node Id] : [dict get $node Name] - [dict get $node Channel] to [dict get $tree Ns] [dict get $tree Nodes $parent_indx Id]"
      }

      set parent_id [dict get $tree Nodes $parent_indx Id]
      set parent_channel [dict get $tree Nodes $parent_indx Channel]

      dict for {child_indx child} [dict get $node Children] {
        copy_node tree $parent_indx subtree $child_indx
      }
    }

    copy_node tree $parent_indx subtree $root_indx
  }

  # ------------------------------------------------------------
  proc print_tree_recursive {tree_ node_indx level} {
    upvar 1 $tree_ tree
    set node [dict get $tree Nodes $node_indx]
    # puts "[string repeat " " $level]+- [dict get $node Id] ([dict get $node Id]) [dict get $node Type]: [dict get $node Channel]"
    set indent [expr {$level + 3}]
    set curAddrDepth [expr {20-$level}]
    puts [format "%*s- %-*s %*s %*s %s" \
            ${indent} "+-"  \
            52 "[dict get $node Name] ([dict get $node Id]) -[dict get $node Type]-" \
            ${curAddrDepth} [dict get $node Address] \
            18 [dict get $node Size] [dict get $node Channel]]
    foreach child_id [dict keys [dict get $node Children]] {
      print_tree_recursive tree $child_id [expr {$level + 1}]
    }
  }

  # ------------------------------------------------------------
  proc print_tree {tree_} {
    upvar 1 $tree_ tree
    dict for {channel_id channel} [dict get $tree Channels] {
      set root_indx [dict get $channel RootNode]
      puts "Channel $channel_id"
      print_tree_recursive tree $root_indx 0
    }
  }

  # ------------------------------------------------------------
  # ------------------------------------------------------------
  namespace export init_tree
  namespace export add_node
  namespace export add_root_node
  namespace export add_child
  namespace export add_subtree
  namespace export print_tree
}
################################################################################
