## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2022 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2022-07-05
# @author Andrea Bellandi andrea.bellandi@desy.de
# --------------------------------------------------------------------------- #
# @brief#
# Part of DESY FPGA Firmware Framework (fwk)
# contains procedures for Xilinx XPS creation and build
# --------------------------------------------------------------------------- #

# ==============================================================================
proc testProject {} {
}

# ==============================================================================
proc cleanProject {PrjBuildName} {
  # delete existing project files if existing
  set curFile "${::fwk::PrjBuildPath}"
  if { [ file exists $curFile ] } {
    file delete -force $curFile
  }
}

# ==============================================================================
proc createProject {PrjBuildName} {
  variable HwFile
  variable MemFile
  variable BitFile
  variable HeaderDirectories ""
  variable AppguruFlags

  if {![info exists ::fwk::CFlags]} {
    namespace eval ::fwk {
      variable CFlags ""
    }
  }

  if {![info exists ::fwk::AccessChannel]} {
    namespace eval ::fwk {
      variable AccessChannel "C8"
    }
  }


  if {[info exists ::fwk::AppguruFlags]} {
    set AppguruFlags $::fwk::AppguruFlags
  } else {
    set AppguruFlags [list]
  }

  ::fwk::printInfo "Creating a Project for Appguru"

  variable verStringSub "\$\{VerString\} $::fwk::VerString"

  set HwFile  [glob [string map $verStringSub $::fwk::HwFile]]
  set BitFile [glob [string map $verStringSub $::fwk::BitFile]]
  set MemFile [glob [string map $verStringSub $::fwk::MemFile]]

  ::fwk::printDivider
  ::fwk::printInfo "Project path : ${::fwk::WorkspacePath}"
  ::fwk::printInfo "Platform name  : ${::fwk::PlatformName}"
  ::fwk::printInfo "HW file: $HwFile"
  ::fwk::printInfo "Mem file: $MemFile"
  ::fwk::printInfo "Bit file: $BitFile"
  ::fwk::printDivider

  # --------------------------------------------------------
  # creating the BSP support

  ::fwk::utils::xps::generateCpuSwBsp $HwFile $::fwk::WorkspacePath $AppguruFlags

}


# ==============================================================================
proc openProject {PrjBuildName} {}

# ==============================================================================
proc closeProject {} {}

# ==============================================================================
proc saveProject {} {

  variable HeaderDirectories
  variable ElfName $::fwk::ReleaseName.elf
  variable ElfPath
  set ElfPath [file join $::fwk::WorkspacePath $ElfName]

  if {![file exists $::fwk::addr::ArtifactsPath/h]} {
    file mkdir $::fwk::addr::ArtifactsPath/h
  }

  lappend HeaderDirectories -I$::fwk::addr::ArtifactsPath/h

  set fp [open [file join $::fwk::WorkspacePath compile.sh] "w"]
  puts $fp "#!/usr/bin/env sh"
  puts $fp "make EXEC=$::fwk::ReleaseName.elf CFLAGS=\"$HeaderDirectories $::fwk::CFlags -DACCESS_CHANNEL_$::fwk::AccessChannel\""
  close $fp

}

# ==============================================================================
proc addSources {args srcList} {

  variable CSrcs [list]
  variable HeaderDirectories

  foreach src $srcList {
    if {[llength $src] >= 2} {
      set path [lindex $src 0]
      set type [lindex $src 1]
      if { $type == "includes"} {
        lappend HeaderDirectories "-I[file normalize $path]"
        ::fwk::printInfo "Includes dir added: [file normalize $path]"
      } elseif { $type == "sources" } {
        set CSrcs [concat $CSrcs [glob -nocomplain -directory $path *.c]]
        set CSrcs [concat $CSrcs [glob -nocomplain -directory $path *.S]]
        set CSrcs [concat $CSrcs [glob -nocomplain -directory $path *.s]]
        ::fwk::printInfo "Source dir added: $path"
      # FWK tries to automatically add type by looking at the file extension
      # if user gave addSources without explicity specifying the type
      } elseif {$type == "h" || $type == "c" } {
        lappend CSrcs $path
      } else {
        ::fwk::printError "type $type not recognized"
        ::fwk::exit -1
      }
    } else {
      lappend CSrcs $src
    }
  }

  foreach csrc $CSrcs {
    file delete -force [file join $::fwk::WorkspacePath [file tail $csrc]]
    file link -symbolic [file join $::fwk::WorkspacePath [file tail $csrc]] $csrc
  }
}

# ==============================================================================
proc buildProject {args} {

  variable ElfName $::fwk::ReleaseName.elf
  variable ElfPath
  set ElfPath [file join $::fwk::WorkspacePath $ElfName]

  ::fwk::printDivider
  ::fwk::printInfo "Compiling $::fwk::ReleaseName.elf"
  ::fwk::printDivider

  set CurDir [pwd]

  cd $::fwk::WorkspacePath

  exec 2>stderr_compile bash compile.sh

  cd $CurDir
}

# ==============================================================================
proc exportOut {} {

  variable ElfPath
  variable BitFile
  variable MemFile

  set CurDir [pwd]
  cd ..

  variable verStringSub "\$\{VerString\} $::fwk::VerString"
  set MemFile [glob [string map $verStringSub $::fwk::MemFile]]
  set BitFile [glob [string map $verStringSub $::fwk::BitFile]]

  ::fwk::releaseFile $ElfPath

  ::fwk::printDivider
  ::fwk::printInfo "Generating ${::fwk::ReleaseName}.bit"
  ::fwk::printDivider

  set ReleasedBit [file join ${::fwk::ReleasePath} ${::fwk::ReleaseName}.bit]
  set stderr_data2mem $::fwk::PrjBuildPath/stderr_data2mem
  exec 2>$stderr_data2mem data2mem -bm $MemFile -bt $BitFile -bd $ElfPath -o b $ReleasedBit

  cd $CurDir
}
