## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2022 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2022-01-09
# @author Lukasz Butkowski <lukasz.butkowski@desy.de>
# --------------------------------------------------------------------------- #
# @brief
# Part of DESY FPGA Firmware Framework (fwk)
# FPGA firmware framework script handling documentation generation
# --------------------------------------------------------------------------- #

namespace eval ::fwk::doc {
  variable pdfThemeDir ${::fwk::ProjectPath}/fwk/tpl/doc
  variable pdfTheme "fwk"
}

# ==============================================================================
# ------------------------------------------------------------------------------
proc ::fwk::doc::main {args} {

  set docTypes [lindex $args 0]

  # check if project do is properly initialized and create initial files if needed
  ::fwk::doc::initPrjDoc

  foreach type $docTypes {
    puts "type: $type"
    switch $type {
      antora {
        ::fwk::doc::genAntoraPlaybook
        ::fwk::doc::genAntora
      }
      pdf {
        ::fwk::doc::genPdf
      }
      adoc {
        ::fwk::addr::main {adoc}
        ::fwk::doc::genAdocReg
      }
      default {
        puts "No doc type specyfied"
      }
    }

  }
}
proc ::fwk::doc::initPrjDoc {} {

  if {![file exists ${::fwk::ProjectPath}/doc/antora.yml]} {
    set fileH [open ${::fwk::ProjectPath}/doc/antora.yml w]
    puts $fileH "name: ${::fwk::ProjectName}"
    puts $fileH "title: ${::fwk::ProjectName}"
    puts $fileH "version: ~"
    puts $fileH "start_page: index.adoc"
    close $fileH
  }
  if {![file exists ${::fwk::ProjectPath}/doc/modules/ROOT/pages]} { file mkdir ${::fwk::ProjectPath}/doc/modules/ROOT/pages}

  if {![file exists ${::fwk::ProjectPath}/doc/modules/ROOT/pages/index.adoc]} {
    set fileH [open ${::fwk::ProjectPath}/doc/modules/ROOT/pages/index.adoc w]
    puts $fileH "= Project ${::fwk::ProjectName} documentation"
    close $fileH
  }
  if {![file exists ${::fwk::ProjectPath}/doc/main.adoc]} {
    set fileH [open ${::fwk::ProjectPath}/doc/main.adoc w]
    puts $fileH "include::modules/ROOT/pages/index.adoc[]"
    close $fileH
  }

}

# ------------------------------------------------------------------------------
proc ::fwk::doc::genPdf {args} {
  ::fwk::printDivider
  ::fwk::printInfo "Generating PDF using asciidoctor-pdf:"
  set mainAdocFile ${::fwk::ProjectPath}/doc/main.adoc
  set cmdToRun \
    [list \
       asciidoctor-pdf \
       -r asciidoctor-diagram \
       -r asciidoctor-mathematical \
       -a mathematical-format=svg \
       -a pdf-themesdir=${::fwk::ProjectPath}/fwk/tpl/doc \
       -a pdf-theme=fwk \
       -o ${::fwk::ProjectPath}/out/${::fwk::ProjectName}_${::fwk::ProjectConf}.pdf \
       $mainAdocFile \
       -a toc \
       -a stem \
       -a toclevels=4 \
       -a sectnums \
       -a text-align=justify \
       -a icons=font \
       -a title-page \
       -a source-highlighter=rouge \
       -a rouge-style=base26.dark \
       -a ProjectName=${::fwk::ProjectName} \
       -a ProjectConf=${::fwk::ProjectConf} \
       -a ProjectVersion=${::fwk::VerString} \
       ]
  puts $cmdToRun
  puts ">>> "

  set dockerImageName "fwfwk/doc-asciidoctor"

  set user [exec id -u]
  set cmdToRunDocker \
    [list \
       docker run -u $user \
       -v ${::fwk::ProjectPath}:/${::fwk::ProjectPath} -w ${::fwk::ProjectPath} --rm \
       -t $dockerImageName ]

  set cmdToRunDocker [concat $cmdToRunDocker $cmdToRun]
  if { [::fwk::utils::existsDockerName $dockerImageName]} {
    ::fwk::printInfo "Found Docker image, generating pdf using $dockerImageName image:"
    if { [catch { eval exec $cmdToRunDocker >@stdout } resulttext] } { puts $resulttext }
  } else {
    if { [catch { eval exec $cmdToRun >@stdout } resulttext] } { puts $resulttext }
  }
  puts ">>>"
  puts "file://${::fwk::ProjectPath}/out/${::fwk::ProjectName}_${::fwk::ProjectConf}.pdf\n"
}

# ------------------------------------------------------------------------------
proc ::fwk::doc::genAntora {args} {
  ::fwk::printDivider
  ::fwk::printInfo "Generating static HTML and PDF using Antora:"
  puts ">>> "
  set AntoraPlaybookFile ${::fwk::PrjBuildPath}/antora_playbook.yml

  set dockerImageName "fwfwk/doc-antora"
  set user [exec id -u]
  set cmdToRun \
    [list \
       antora --stacktrace  $AntoraPlaybookFile
      ]

  set cmdToRunDocker \
    [list \
       docker run -u $user \
       -v ${::fwk::ProjectPath}:/${::fwk::ProjectPath} -w ${::fwk::ProjectPath} --rm \
       -t $dockerImageName ]
  set cmdToRunDocker [concat $cmdToRunDocker $cmdToRun]

  if { [::fwk::utils::existsDockerName $dockerImageName]} {
    ::fwk::printInfo "Found Docker image, generating antora using $dockerImageName image:"
    if { [catch { eval exec $cmdToRunDocker >@stdout } resulttext] } { puts $resulttext }
  } else {
    if { [catch { eval exec $cmdToRun >@stdout } resulttext] } { puts $resulttext }
  }
  puts ">>>"
  puts "Genrated Anrtora PDF file."
  puts "file://${::fwk::ProjectPath}/out/${::fwk::ProjectName}_${::fwk::ProjectConf}/doc/pdf/"
  puts ">>>"
}


# ------------------------------------------------------------------------------
proc ::fwk::doc::genAntoraPlaybook {} {

  set AntoraPlaybookFile ${::fwk::PrjBuildPath}/antora_playbook.yml
  set antoraPlaybook [open $AntoraPlaybookFile w]

  puts $antoraPlaybook "runtime:"
  puts $antoraPlaybook "  cache_dir: ${::fwk::PrjBuildPath}/.antoracache"
  puts $antoraPlaybook "site:"
  puts $antoraPlaybook "  title: ${::fwk::ProjectName}_${::fwk::ProjectConf} ${::fwk::VerString}"
  puts $antoraPlaybook "  url: /"
  puts $antoraPlaybook "  start_page: ${::fwk::ProjectName}::index.adoc"
  puts $antoraPlaybook "asciidoc:"
  puts $antoraPlaybook "  attributes:"
  puts $antoraPlaybook "    idprefix: ''"
  puts $antoraPlaybook "    idseparator: '-'"
  puts $antoraPlaybook "    linkattrs: ''"
  puts $antoraPlaybook "    toc: ~"
  puts $antoraPlaybook "    page-toclevels: 3@"
  puts $antoraPlaybook "    source-highlighter: ~" ;# enabled with supplemental UI, default has no VHDL, Tcl, etc.
  puts $antoraPlaybook "    kroki-fetch-diagram: true"
  puts $antoraPlaybook "    stem: '@'"
  puts $antoraPlaybook "    page-pagination: ''"
  puts $antoraPlaybook "    hide-uri-scheme: '@'"
  puts $antoraPlaybook "    icons: font"
  puts $antoraPlaybook "    xrefstyle: short"
  puts $antoraPlaybook "    ProjectName: ${::fwk::ProjectName}"
  puts $antoraPlaybook "    ProjectConf: ${::fwk::ProjectConf}"
  puts $antoraPlaybook "    ProjectVersion: ${::fwk::VerString}"
  puts $antoraPlaybook "  extensions:"
  puts $antoraPlaybook "    - asciidoctor-kroki"
  puts $antoraPlaybook "    - '@djencks/asciidoctor-mathjax'"
  puts $antoraPlaybook "antora:"
  puts $antoraPlaybook "  extensions:"
  puts $antoraPlaybook "    - '@antora/lunr-extension'"
  puts $antoraPlaybook "    - '@antora/collector-extension'"
  puts $antoraPlaybook "    - '@antora/pdf-extension'"
  puts $antoraPlaybook "ui:"
  puts $antoraPlaybook "  bundle:"
  puts $antoraPlaybook "    url: https://gitlab.com/antora/antora-ui-default/-/jobs/artifacts/HEAD/raw/build/ui-bundle.zip?job=bundle-stable"
  puts $antoraPlaybook "    snapshot: true"
  puts $antoraPlaybook "  supplemental_files: ${::fwk::ProjectPath}/fwk/tpl/doc/antora_ui"
  puts $antoraPlaybook "output:"
  puts $antoraPlaybook "  dir: ${::fwk::ProjectPath}/out/${::fwk::ProjectName}_${::fwk::ProjectConf}/doc/html"
  puts $antoraPlaybook "content:"
  puts $antoraPlaybook "  sources:"
  puts $antoraPlaybook "    - url: ${::fwk::ProjectPath}"
  puts $antoraPlaybook "      branches: HEAD"
  # puts $antoraPlaybook "      start_path: doc"
  puts $antoraPlaybook "      start_paths:"
  puts $antoraPlaybook "      - doc"

  #set nl [::fwk::createNamespacesList ::fwk::src]
  set nl [::fwk::utils::listns ::fwk::src]
  set modPaths {}
  foreach module $nl {
    lappend modPaths [subst $${module}::Path ]
  }
  set smodPaths [lsort -unique $modPaths]

  foreach modPath $smodPaths {
    puts "Adding repo of $module"
    set repoPathLen [string length ${::fwk::ProjectPath}]
    set start_path "[string range $modPath [expr {$repoPathLen+1}] end]/doc"
    if {[file exists $modPath/doc/antora.yml]} {
      puts $antoraPlaybook "      - $start_path"
    }

  }

  close $antoraPlaybook

  set AntoraPlaybookFile ${::fwk::PrjBuildPath}/antora-assembler.yml
  set antoraPlaybook [open $AntoraPlaybookFile w]
  puts $antoraPlaybook "root_level: 0"
  puts $antoraPlaybook "component_versions: '*'"
  puts $antoraPlaybook "asciidoc:"
  puts $antoraPlaybook "  attributes:"
  puts $antoraPlaybook "    allow-uri-read: ''"
  puts $antoraPlaybook "    kroki-fetch-diagram: ~"
  puts $antoraPlaybook "    stem: ''"
  puts $antoraPlaybook "    icons: font"
  puts $antoraPlaybook "    idprefix: ''"
  puts $antoraPlaybook "    idseparator: '-'"
  puts $antoraPlaybook "    linkattrs: ''"
  puts $antoraPlaybook "    toc: ''"
  puts $antoraPlaybook "    toclevels: 4"
  puts $antoraPlaybook "    sectnums: ''"
  puts $antoraPlaybook "    text-align: justify"
  puts $antoraPlaybook "    source-highlighter: rouge"
  puts $antoraPlaybook "    rouge-style: base26.dark"
  puts $antoraPlaybook "    ProjectName: ${::fwk::ProjectName}"
  puts $antoraPlaybook "    ProjectConf: ${::fwk::ProjectConf}"
  puts $antoraPlaybook "    ProjectVersion: ${::fwk::VerString}"
  puts $antoraPlaybook "    pdf-themesdir: ${::fwk::doc::pdfThemeDir}"
  puts $antoraPlaybook "    pdf-theme: ${::fwk::doc::pdfTheme}"
  puts $antoraPlaybook "    mathematical-format: svg"
  puts $antoraPlaybook "build:"
  puts $antoraPlaybook "  command: asciidoctor-pdf -r asciidoctor-mathematical -r asciidoctor-kroki"
  puts $antoraPlaybook "  dir: ${::fwk::ProjectPath}/out/${::fwk::ProjectName}_${::fwk::ProjectConf}/doc/pdf"

  close $antoraPlaybook
}


proc ::fwk::doc::genAdocReg {} {

  #
  set topFile [open $::fwk::addr::ArtifactsPath/adoc/addressmap.adoc w]
  puts $topFile "= Address Space"
  puts $topFile ":toc:"
  puts $topFile ":page-toclevels: 4"
  puts $topFile "\n\n"
  puts $topFile "== Tree"
  dict for {accessChannel nodes} $::fwk::AddressSpaceChDict {
    puts $topFile ".Address Tree $accessChannel "
    puts $topFile "\[plantuml, format=svg\]"
    puts $topFile "...."
    puts $topFile "@startmindmap"
    puts $topFile "skinparam monochrome true"
    puts $topFile "skinparam ranksep 20"
    puts $topFile "skinparam arrowThickness 0.7"
    puts $topFile "skinparam packageTitleAlignment left"
    puts $topFile "skinparam usecaseBorderThickness 0.4"
    puts $topFile "skinparam defaultFontSize 12"
    puts $topFile "skinparam rectangleBorderThickness 1"
    puts $topFile "* Access channel $accessChannel"
    foreach nodeId [dict keys $nodes] {
      set instanceName [dict get $nodes $nodeId Id]
      set name [dict get $nodes $nodeId Name]
      set depth [expr [dict get $nodes $nodeId Depth] + 0 ]
      if { $instanceName != "" } {
        puts $topFile "[string repeat * $depth] ${instanceName} ($name)"
      }
    }
    puts $topFile "@endmindmap"
    puts $topFile "...."
  }
  dict for {accessChannel nodes} $::fwk::AddressSpaceChDict {
    # puts $topFile "\nxref:ch_${accessChannel}_address_map.adoc\[Access channel $accessChannel\]\n"
    puts $topFile "\ninclude::ch_${accessChannel}_address_map.adoc\[Access channel $accessChannel,leveloffset=1\]\n"
    foreach nodeId [dict keys $nodes] {
    }
  }

  close $topFile

  dict for {accessChannel nodes} $::fwk::AddressSpaceChDict {
    set chFile [open $::fwk::addr::ArtifactsPath/adoc/ch_${accessChannel}_address_map.adoc w]
    puts $chFile "= Access channel $accessChannel\n"
    foreach nodeId [dict keys $nodes] {
      set instanceName [dict get $nodes $nodeId Id]
      set depth [expr [dict get $nodes $nodeId Depth] - 1 ]
      # set depth [dict get $nodes $nodeId Depth]
      # if virtual top do not palce in doc
      if { $instanceName != "" } {
        puts $chFile "include::addrmap_${instanceName}.adoc\[leveloffset=\+$depth]\n"
      }

    }
    close $chFile
  }
  if {![file exists $::fwk::ProjectPath/doc/modules/addressmap_$::fwk::ProjectConf/pages ]} {
    file mkdir $::fwk::ProjectPath/doc/modules/addressmap_$::fwk::ProjectConf/pages}
  file delete -force $::fwk::ProjectPath/doc/modules/addressmap_$::fwk::ProjectConf/pages
  file copy -force $::fwk::addr::ArtifactsPath/adoc/ $::fwk::ProjectPath/doc/modules/addressmap_$::fwk::ProjectConf/pages
}

