## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2021 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2021-12-06
# @author Lukasz Butkowski  <lukasz.butkowski@desy.de>
# --------------------------------------------------------------------------- #
# @brief
# Part of DESY FPGA Firmware Framework (fwk)
# Contains procedures for Xilinx PlanAhead project creation and build
# --------------------------------------------------------------------------- #

# ==============================================================================
proc testProject {} {
}

# ==============================================================================
proc cleanProject {PrjBuildName} {
  # delete existing project files if existing
  set ProjExt [ list ppr data gen]
  foreach ext $ProjExt {
    set curFile "${PrjBuildName}.$ext"
    if { [ file exists $curFile ] } {
      file delete -force $curFile
    }
  }

}

# ==============================================================================
proc createProject {PrjBuildName} {
  create_project -force $PrjBuildName
}

# ==============================================================================
proc openProject {PrjBuildName} {

  set Prj ${PrjBuildName}.ppr
  if { ! [ file exists $Prj ] } {
    ## project file isn't there, rebuild it.
    ::fwk::printDivider
    ::fwk::printError "Project ${::fwk::PrjBuildName} not found. Use create command to recreate it."
    ::fwk::exit -1
  }
  if {[set result [catch {open_project $Prj} resulttext]]} {
    ::fwk::printInfo "Project already opened."
  }
  update_compile_order -fileset [current_fileset]
}

# ==============================================================================
proc closeProject {} {
  close_project
}

# ==============================================================================
proc saveProject {} {}

# ==============================================================================
proc addSources {srcDictList} {

  # get file list
  set srcFiles {}
  foreach srcDict $srcDictList {
    lappend srcFiles [dict get $srcDict File]
  }

  # add files # TODO: optional tool args key
  # if { $args == "" } {
  if {[set result [catch {eval add_files $srcFiles} resulttext]]} {
    ::fwk::printDivider
    ::fwk::utils::cprint $resulttext {"ERROR" -fg red -style bright}
    ::fwk::printError "PlanAhead `add_files` failed"
    ::fwk::exit -2
  }
  # } else {
  #   if {[set result [catch {eval add_files $args $srcFiles} resulttext]]} {
  #     ::fwk::printDivider
  #     ::fwk::utils::cprint $resulttext {"ERROR" -fg red -style bright}
  #     ::fwk::printError "add_files failed"
  #     ::fwk::exit -2
  #   }
  # }

  # set files properties
  foreach srcDict $srcDictList {
    set srcFile   [dict get $srcDict File]
    set srcType   [dict get $srcDict Type]
    set srcLib    [dict get $srcDict Library]
    # optional keys
    set usedIn ""
    set fileset ""
    set procOrder ""
    if {[dict exists $srcDict UsedIn]} {
      set usedIn    [dict get $srcDict UsedIn] }
    if {[dict exists $srcDict ProcOrder]} {
      set procOrder [dict get $srcDict ProcOrder] }
    if {[dict exists $srcDict Fileset]} {
      set fileset   [dict get $srcDict Fileset] }
    # support upper and lower case in the list or different format
    switch -regexp [string tolower $srcType] {
      "vhdl?\$" {
        set_property FILE_TYPE "VHDL" [get_files $srcFile]
        set_property LIBRARY $srcLib [get_files $srcFile]
        if { $usedIn != "" } {
          if {[string compare -nocase $usedIn "simulation"] == 0} {
            # remove the sim file from 'sources_1' to place it in proper simulation fileset
            remove_files $srcFile
            # add file to sim fileset (if $srcLib is set, create corresponding fileset; otherwise use the default sim_1)
            if { $fileset != "" }    {
              create_fileset -simset -quiet $fileset
              add_files -fileset $fileset -norecurse $srcFile
            } else {
              add_files -fileset sim_1 -norecurse $srcFile
            }
            set_property LIBRARY $srcLib [get_files $srcFile]
          }
          set_property FILE_TYPE "VHDL" [get_files $srcFile]
        }
      }
      "v(erilog)?\$" {
        set_property FILE_TYPE "Verilog" [get_files $srcFile]
        set_property LIBRARY $srcLib [get_files $srcFile]
        if { $usedIn != "" } {
          if {[string compare -nocase $usedIn "simulation"] == 0} {
            # remove the sim file from 'sources_1' to place it in proper simulation fileset
            remove_files $srcFile
            # add file to sim fileset (if $srcLib is set, create corresponding fileset; otherwise use the default sim_1)
            if { $fileset != "" }    {
              create_fileset -simset -quiet $fileset
              add_files -fileset $fileset -norecurse $srcFile
            } else {
              add_files -fileset sim_1 -norecurse $srcFile
            }
            set_property LIBRARY $srcLib [get_files $srcFile]
          }
          set_property FILE_TYPE "VHDL" [get_files $srcFile]
        }
      }
      "vhdl? ?(20)?08\$" {
        set_property FILE_TYPE "VHDL" [get_files $srcFile]
        set_property LIBRARY $srcLib [get_files $srcFile]
        if { $usedIn != "" } {
          if {[string compare -nocase $usedIn "simulation"] == 0} {
            # remove the sim file from 'sources_1' to place it in proper simulation fileset
            remove_files $srcFile
            # add file to sim fileset (if $srcLib is set, create corresponding fileset; otherwise use the default sim_1)
            if { $fileset != "" }    {
              create_fileset -simset -quiet $fileset
              add_files -fileset $fileset -norecurse $srcFile
            } else {
              add_files -fileset sim_1 -norecurse $srcFile
            }
            set_property LIBRARY $srcLib [get_files $srcFile]
          }
          set_property FILE_TYPE "VHDL" [get_files $srcFile]
        }
        # if { $usedIn != "" } { set_property USED_IN $usedIn [get_files $srcFile] }
        puts "# WARNING: VHDL 2008 type not supported in PlanAhead"
      }
      "ucf\$" {
        remove_files $srcFile;
        # add file to constraints fileset (if $srcLib is set, create corresponding fileset; otherwise use the default constrs_1)
        if { $fileset != "" }    {
          create_fileset -constrset -quiet $fileset
          add_files -fileset $fileset -norecurse $srcFile
        } else {
          add_files -fileset constrs_1 -norecurse $srcFile
        }
        set_property FILE_TYPE "UCF" [get_files $srcFile]
      }
      default {
        #do nothing
      }
    }
  }

}

# ==============================================================================
proc buildProject {args} {
  # ------------------------------------------------------------------
  # compilation
  reset_run "impl_1"
  reset_run "synth_1"

  if {[catch { launch_runs synth_1 } resulttext ]} { puts $resulttext; ::fwk::exit -1 }
  wait_on_run synth_1

  if {[catch { launch_runs impl_1 } resulttext ]} {
    puts "Synth report:"
    set grepResult [::fwk::utils::grepFiles "ERROR:" [glob ${::fwk::PrjBuildPath}/${::fwk::PrjBuildName}.runs/synth_1/*.srp]]
    puts [::fwk::utils::cgrep $grepResult "ERROR" -fg red -style bright]

    puts "$resulttext"; ::fwk::exit -1 }
  if {[catch { wait_on_run impl_1 } resulttext ]} { ::fwk::exit -1 }

  catch { launch_runs impl_1 -to_step Bitgen }
  wait_on_run [current_run]

}


# ==============================================================================
#! export bit file to out folder
proc exportOut {} {

  set top [get_property TOP [current_fileset]]
  set dir [get_property DIRECTORY [current_run]]
  puts "Copy bit file to: $::fwk::ProjectPath/out/${::fwk::PrjBuildName}/${::fwk::PrjBuildName}_${::fwk::VerString}.bit"
  if {[catch { file copy -force ${dir}/${top}.bit $::fwk::ProjectPath/out/${::fwk::PrjBuildName}/${::fwk::PrjBuildName}_${::fwk::VerString}.bit } resulttext ]} {
    puts "$resulttext";
    ::fwk::exit -1
  }
}

# ==============================================================================
#! function add .ngc ipcore file,if not exist it is trying to regenerate xco ipcore
proc genAddIP { IpPath } {

}

# ==============================================================================
# dummy
proc addGenIPs { args sources} {}
proc packageIp {args} {}
proc addIp {args} {}
proc wrapIP {args} {}
