## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2022 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2022-04-27
# @author Seyed Nima Omidsajedi <nima.sajedi@desy.de>
# @author Lukasz Butkowski <lukasz.butkowski@desy.de>
# --------------------------------------------------------------------------- #
# @brief#
# Part of DESY FPGA Firmware Framework (fwk)
# contains procedures for Xilinx Vitis creation and build
# --------------------------------------------------------------------------- #

# ==============================================================================
proc testProject {} {
}

# ==============================================================================
proc cleanProject {PrjBuildName} {
  # delete existing project files if existing
  set curFile "${::fwk::PrjBuildPath}"
  if { [ file exists $curFile ] } {
    file delete -force $curFile
  }
}

# ==============================================================================
proc createProject {PrjBuildName} {
  variable ::fwk::HwFile
  #variable AppName
  ::fwk::printInfo "Create Project from fwk_vitis.tcl, createProject proc"
  ::fwk::printInfo "Create Vitis Platform"

  # check variables HW file
  if { ![info exists ::fwk::HwFile] } {
    ::fwk::printError "No ::fwk::HwFile defined. Please set env variable FWK_HW_FILE. e.g. export FWK_HW_FILE=top.xsa"; ::fwk::exit -2;}
  if { ![info exists ::fwk::CpuType] } {
    ::fwk::printError "No ::fwk::CpuType defined. Please set CpuType in cfg/config.cfg e.g. CpuType=psu_cortexa53_0";  ::fwk::exit -2; }
  if { ![info exists ::fwk::Arch] } {
    ::fwk::printError "No ::fwk::Arch defined. Please set Arch in cfg/config.cfg e.g. Arch=64-bit";  ::fwk::exit -2; }
  if { ![info exists ::fwk::OsType] } {
    ::fwk::printError "No ::fwk::OsType defined. Please set OsType in cfg/config.cfg e.g. OsType=standalone";  ::fwk::exit -2; }

  if { ![info exists ::fwk::AppName] } {
    ::fwk::printWarning "No ::fwk::AppName defined. An empty template will be created. It can later be copied to another location. Set AppName in cfg/name_of_sw_project.cfg to the folder name within ::fwk::SrcPath, e.g. AppName=hello_world"
  } elseif { ![info exists ::fwk::AppLang] } {
    ::fwk::printWarning "An application was specified but no ::fwk::AppLang defined. Using C by default. Set AppLang in cfg/name_of_sw_project.cfg to either \"c\" or \"c++\", e.g. AppName=c++"
  }

  set appLang "c"
  if { [info exists ::fwk::AppLang] } {
    if { [llength $::fwk::AppLang] > 0} {
      set appLang $::fwk::AppLang
    }
  }

  setws -switch ${::fwk::WorkspacePath}
  ::fwk::printDivider
  ::fwk::printInfo "Workspace path : ${::fwk::WorkspacePath}"
  ::fwk::printInfo "Platform name  : ${::fwk::PlatformName}"
  ::fwk::printInfo "HW file: ${::fwk::HwFile}"
  ::fwk::printDivider

  # get Vitis version
  set xsctVersion [lindex [version] 1]

  ::fwk::printInfo "Xilinx Vitis Version $xsctVersion is used"
  # --------------------------------------------------------
  # create platform
  ::fwk::printInfo "Creating Plarform..."
  if {[set result [catch {eval \
                            platform create \
                            -name ${::fwk::PlatformName} \
                            -hw   ${::fwk::HwFile} \
                            -arch ${::fwk::Arch} \
                            -fsbl-target ${::fwk::CpuType}
  } resulttext]]} {
    puts ""
    ::fwk::printError "$resulttext"
    ::fwk::exit -2
  }

  platform write

  ::fwk::printInfo "Creating Domain..."
  domain create -name {app_domain} -os ${::fwk::OsType} -proc ${::fwk::CpuType} -runtime {cpp} -support-app {empty_application}

  platform write
  platform active ${::fwk::PlatformName}

  if { [ info exists ::fwk::Stdin ] } {
    bsp config stdin ${::fwk::Stdin}
  }
  if { [ info exists ::fwk::Stdout ] } {
    bsp config stdout ${::fwk::Stdout}
  }
  bsp write
  bsp reload
  bsp regenerate

  platform generate -quick

  if { [info exists ::fwk::AppName] } {
    # --------------------------------------------------------
    # creating application on a platform domain
    ::fwk::printCBM "\nCreating Vitis Application"

    set targetDir "${::fwk::WorkspacePath}/${::fwk::AppName}/src"
    file mkdir ${::fwk::WorkspacePath}/${::fwk::AppName}

    if { [ file exists $targetDir ] } {
      file delete -force $targetDir
      ::fwk::printWarning "Vitis application: Overwrite to existing target directory"
    }

    if { $appLang == "c" } {
      if { $xsctVersion < 2021.1 } {
        app create -name ${::fwk::AppName} -domain "app_domain" -template "Empty Application" -proc ${::fwk::CpuType} -lang $appLang
      } else {
        app create -name ${::fwk::AppName} -domain "app_domain" -template "Empty Application(C)" -proc ${::fwk::CpuType} -lang $appLang
      }
    } elseif { $appLang == "c++" } {
      app create -name ${::fwk::AppName} -domain "app_domain" -template "Empty Application (C++)" -proc ${::fwk::CpuType} -lang $appLang
    }

    # DesyRDL-generated headers are being placed here, so add it to the include path
    app config -name ${::fwk::AppName} include-path ${::fwk::WorkspacePath}/include

    # Remove the linker script that Vitis auto-generates; we have our own.
    file delete -force $targetDir
  }
}


# ==============================================================================
proc openProject {PrjBuildName} {
  setws -switch $::fwk::WorkspacePath

  if { ! [ file exists $::fwk::WorkspacePath ] } {
    ## project file isn't there, rebuild it.
    ::fwk::printError "Project ${::fwk::PrjBuildName} not found. Use create command to recreate it."
    ::fwk::exit -1
  }

  if { ![info exists ::fwk::AppName] } {
    ::fwk::printWarning "No ::fwk::AppName defined. Please set AppName in cfg/config.cfg e.g. AppName=AppName. Using project name"
    set ::fwk::AppName $::fwk::ProjectName
  }

}

# ==============================================================================
proc closeProject {} {
}

# ==============================================================================
proc saveProject {} {}

# ==============================================================================
proc addSources {args srcList} {

  foreach src $srcList {
    set path [lindex $src 0]
    set type [lindex $src 1]

    set appDir ${::fwk::WorkspacePath}/${::fwk::AppName}
    set absPath [file normalize $path]

    if { $type == "includes"} {
      if { ![file isdirectory $path] } {
        ::fwk::printError "fwk can only add directories of include files to a Vitis project, not single files. Cannot add $path."
        ::fwk::exit -1
      }
      app config -name ${::fwk::AppName} include-path $absPath

    } elseif { $type == "sources"} {
      if { ![file isdirectory $path] } {
        ::fwk::printError "fwk can only add directories of source files to a Vitis project, not single files. Cannot add $path."
        ::fwk::exit -1
      }

      # we want a symlink that replicates the source path relative to ::fwk::SrcPath
      set len1 [string length $::fwk::SrcPath]
      set len2 [string length $absPath]
      set srcSubdirPath [string range $absPath $len1 $len2]

      set targetDir $appDir/src/$srcSubdirPath

      # make sure parent folder exists before creating the symlink
      file mkdir [file dirname $targetDir]
      file link -symbolic $targetDir $absPath

    } elseif { $type == "linker-script"} {
      # The application project is created with the "Empty Application" template.
      # This template includes a generated linker script.
      # Replace it with a symlink at the same location.
      file link -symbolic $appDir/src/[file tail $path] $absPath
    }

  }
}

# ==============================================================================
proc buildProject {args} {

  setws -switch ${::fwk::WorkspacePath}

  platform active ${::fwk::PlatformName}
  domain active {app_domain}

  if {[set result [catch {eval \
                            app build -name ${::fwk::AppName}
  } resulttext]]} {
    puts ""
    ::fwk::printError "$resulttext"
    ::fwk::exit -2
  }
  puts ""
  ::fwk::printCBM "== Report Application and its Domain =="

  set app_report [app report -name ${::fwk::AppName}]
  set domain_report [domain report -name app_domain]
  puts $app_report
  puts $domain_report
}

# ==============================================================================
proc exportOut {} {
  ::fwk::printInfo "Copy elf file of the project to artifacts out"
  puts "from:  ${::fwk::WorkspacePath}/${::fwk::AppName}/Debug/${::fwk::AppName}.elf"
  puts "to  :  $::fwk::ProjectPath/out/${::fwk::PrjBuildName}/${::fwk::AppName}.elf"

  if { [catch {file copy -force \
                 ${::fwk::WorkspacePath}/${::fwk::AppName}/Debug/${::fwk::AppName}.elf \
                 $::fwk::ProjectPath/out/${::fwk::PrjBuildName}/${::fwk::AppName}.elf} resulttext ] } {
    puts ""
    ::fwk::printError "$resulttext"
    ::fwk::printInfo "For more information check the project build log !"

    ::fwk::exit -2
  }
}
