## -------------------------------------------------------------------------- #
#           ____  _____________  __                                           #
#          / __ \/ ____/ ___/\ \/ /                 _   _   _                 #
#         / / / / __/  \__ \  \  /                 / \ / \ / \                #
#        / /_/ / /___ ___/ /  / /               = ( M | S | K )=              #
#       /_____/_____//____/  /_/                   \_/ \_/ \_/                #
#                                                                             #
# --------------------------------------------------------------------------- #
# @copyright Copyright 2019-2023 DESY
# SPDX-License-Identifier: Apache-2.0
# --------------------------------------------------------------------------- #
# @date 2023-01-26
# @author Lukasz Butkowski  <lukasz.butkowski@desy.de>
# @author Seyed Nima Omidsajedi <nima.sajedi@desy.de>
# --------------------------------------------------------------------------- #
# @brief
# Part of DESY FPGA Firmware Framework (fwk)
# Main Tcl script for the firmware framework. Starting point.
# --------------------------------------------------------------------------- #

namespace eval ::fwk {}

global ExecutionPath
global ColorTerminal
global Exit
set Exit 0
set ColorTerminal 0

# -----------------------------------------------------------------------------
# main
proc ::fwk::main {args} {

  global ColorTerminal
  global ScriptPath
  global ExecutionPath
  global Exit

  # set lang to be used
  set ::env(LANG) "en_US.UTF-8"
  set ::env(LANGUAGE) "en_US.UTF-8"
  set ::env(LC_ALL) "en_US.UTF-8"

  set dispScriptFile [file normalize [info script]]
  set ScriptPath [file dirname $dispScriptFile]
  set ExecutionPath [pwd]

  set ::fwk::GuiMode 0 ;       # default no gui
  set ::fwk::ShellMode 0 ;     # default no shell

  set ::fwk::ModuleOwnLib 0 ;  # default Modules in common default libraries

  # if rerun on reinit, Path already exist, just cd
  if {[info exists ::fwk::ProjectPath] } {
    cd $::fwk::ProjectPath/fwk/src
  } else {
    # go to script location
    # puts "root: [file tail $dispScriptFile]"
    if { [file tail $dispScriptFile] == "vsim"} {
      set args [lrange [lindex $args end] 2 end]
      # redefine exit procedure for Modelsim
      proc ::fwk::exit {{arg}} { if { $arg != "" } {quit -code $arg} else {quit}}
      cd fwk/src
    } else {
      cd $ScriptPath
    }

    # set absolute path of the current project
    set ::fwk::ProjectPath [file normalize "../../"]
    set ::fwk::Path ${::fwk::ProjectPath}/fwk/src
  }
  # create tool type variable if does not exist
  if {![info exists ::fwk::ToolType] } {set ::fwk::ToolType "" }

  # -----------------------------------------------------------------------------
  # load libraries
  # load dict library for Tcl version 8.4 (e.g. Xilinx ISE)
  set TclVersion [info patchlevel]
  set match [regexp {(\d+)\.(\d+).(\d+)} $TclVersion totalmach TclMajor TclMinor TclPatch]
  if { "$TclMajor.$TclMinor" == "8.4" } {
    sourceFwkFile $::fwk::Path/dict.tcl
  }

  # sourceFwkFile main framework
  sourceFwkFile fwk.tcl

  # sourceFwkFile fwk utils library
  sourceFwkFile fwk_utils.tcl

  # sourceFwkFile fwk address library
  sourceFwkFile fwk_addr.tcl

  # sourceFwkFile fwk doc library
  sourceFwkFile fwk_doc.tcl

  # default variables settings
  set taskoption ""
  set toolType ""
  set AddressType ""
  set Exit 0
  set ConfigFile "../../cfg/default.cfg"

  # get noexit option arg
  set args [::fwk::utils::parseArgValue $args "--colorterm" ColorTerminal]

  ::fwk::printDivider
  puts ""
  ::fwk::printCBM "    __\/\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\__\/\\\\\\______________\/\\\\\\__\/\\\\\\________\/\\\\_"
  ::fwk::printCBM "     _\\\/\\\\\\\/\/\/\/\/\/\/\/\/\/\/__\\\/\\\\\\_____________\\\/\\\\\\_\\\/\\\\\\_____\/\\\\\/\/__"
  ::fwk::printCBM "      _\\\/\\\\\\_____________\\\/\\\\\\_____________\\\/\\\\\\_\\\/\\\\\\__\/\\\\\\\/\/_____"
  ::fwk::printCBM "       _\\\/\\\\\\\\\\\\\\\\\\\\\\_____\\\/\/\\\\\\____\/\\\\\\____\/\\\\\\__\\\/\\\\\\\\\\\\\/\/\\\\\\_____"
  ::fwk::printCBM "        _\\\/\\\\\\\/\/\/\/\/\/\/_______\\\/\/\\\\\\__\/\\\\\\\\\\__\/\\\\\\___\\\/\\\\\\\/\/_\\\/\/\\\\\\____"
  ::fwk::printCBM "         _\\\/\\\\\\_______________\\\/\/\\\\\\\/\\\\\\\/\\\\\\\/\\\\\\____\\\/\\\\\\____\\\/\/\\\\\\___"
  ::fwk::printCBM "          _\\\/\\\\\\________________\\\/\/\\\\\\\\\\\\\/\/\\\\\\\\\\_____\\\/\\\\\\_____\\\/\/\\\\\\__"
  ::fwk::printCBM "           _\\\/\\\\\\_________________\\\/\/\\\\\\__\\\/\/\\\\\\______\\\/\\\\\\______\\\/\/\\\\\\_"
  ::fwk::printCBM "            _\\\/\/\/___________________\\\/\/\/____\\\/\/\/_______\\\/\/\/________\\\/\/\/__"
  puts ""
  ::fwk::printDivider
  ::fwk::printCBM " FPGA Firmware Framework Project (FWK)"
  ::fwk::printCBM "                                                 developed by MSK Group of DESY"
  ::fwk::printDivider

  ::fwk::printInfo "Absolute path of the firmware framework project:\n\[::fwk::ProjectPath\] ${::fwk::ProjectPath}"

  #puts "Arguments:"
  ::fwk::printInfo "main.tcl \[args\]: $args"

  if { [llength $args] == 0 } {
    return
  }

  # get noexit option arg
  set args [::fwk::utils::parseArgFlag $args "--exit" Exit]
  set ::fwk::ExitMode $Exit

  # get project config file arg
  set args [::fwk::utils::parseArgValue $args "-c" ConfigFile]

  # get tool
  set args [::fwk::utils::parseArgValue $args "-t" toolType]

  ::fwk::loadProjectInfo $ConfigFile

  if { $toolType != $::fwk::ToolType && $toolType != "" } {
    ::fwk::printWarning "Overwriting configuration tool type with $toolType\n"
    set ::fwk::ToolType $toolType
    unset toolType
  }

  # sourceFwkFile the tool backend
  ::fwk::printInfo "Sourcing  fwk_${::fwk::ToolType}.tcl"
  if { [file exists ${::fwk::Path}/fwk_${::fwk::ToolType}.tcl] } {
    namespace eval ::fwk::tool {
      # indicates whether the tool has tool-specific procedures
      variable SetPrjProperties 1
      sourceFwkFile ${::fwk::Path}/fwk_${::fwk::ToolType}.tcl
      namespace export createProject buildProject postBuildProject cleanProject
      namespace export genAddIP exportOut addSources
    }
  } else {
    ::fwk::printWarning "no tool fwk_${::fwk::ToolType}.tcl backend found.\n"
  }

  # sourceFwkFile optional libraries if proper variables set
  # OSVVM
  if { [info exists ::env(FWK_OSVVM_PATH)] } {
    if { ! [file exists $::env(FWK_OSVVM_PATH)]} {
      ::fwk::printWarning "FWK_OSVVM_PATH: $::env(FWK_OSVVM_PATH) path does not exists";
    } else {
      set ::fwk::OsvvmPath $::env(FWK_OSVVM_PATH)
      ::fwk::printInfo "Sourcing OSVVM scripts from ::env(FWK_OSVVM_PATH): $::fwk::OsvvmPath"
      namespace eval ::osvvmfwk {
        sourceFwkFile $::env(FWK_OSVVM_PATH)/Scripts/StartUp.tcl
      }
    }
  }

  # create out folder if does not exist
  ::fwk::createOutDir

  # --------------------------------------------
  set runResult [eval ::fwk::run $args]
  # --------------------------------------------

  # exit tool
  if { $Exit == 1 } {
    puts "Exiting tool..."
    if { [file tail $dispScriptFile] == "vsim"} {quit -sim}; # when vsim quit simulation first
    cd $ExecutionPath
    exit $runResult
  }

}

# -----------------------------------------------------------------------------
proc ::fwk::run { args } {

  variable taskOption {}
  variable addressType {}
  variable docType {}

  # get addrestype option args
  set args [::fwk::utils::parseArgValue $args "-a" addressType]

  # get doc type
  set args [::fwk::utils::parseArgValue $args "-dt" docType]

  # get task option arg
  set args [::fwk::utils::parseArgValue $args "-o" taskOption]

  ::fwk::getVersion ::fwk
  ::fwk::printProjectInfo
  ::fwk::listSubmodules

  set ::fwk::ReleasePath  [file join $::fwk::ProjectPath out $::fwk::PrjBuildName]
  set ::fwk::ReleaseName  ${::fwk::PrjBuildName}_${::fwk::VerString}

  foreach task $args {
    switch $task {
      "info"   {
        # default prints just initial
      }
      "init" {
        if { [namespace exists ::fwk::src] } { namespace delete ::fwk::src }
        if { [namespace exists ::fwk::ip] } { namespace delete ::fwk::ip }
        if { [namespace exists ::fwk::lib] } { namespace delete ::fwk::lib }
        ::fwk::printCBM "\n== INIT =="
        # if { [info exists ::fwk::OsvvmPath]} {
        #   ::fwk::addModule ::fwk::lib osvvm ./lib_osvvm.tcl
        # } else {                # evaluate and create empty namespace
          namespace eval ::fwk::lib {}
        # }
        ::fwk::addModule ::fwk::src [dict create \
                            InstanceName "" \
                            File $::fwk::ProjectTcl \
                            Type "Project" \
                           ]

        ::fwk::genSrcTree
      }
      "addr"   {
        ::fwk::printCBM "\n== ADDRESS SPACE =="
        ::fwk::addr::main $addressType
      }
      "doc"   {
        ::fwk::printCBM "\n== DOCUMENTATION =="
        ::fwk::doc::main $docType
      }
      "shell"     { ::fwk::openProject ;  set ::fwk::ShellMode 1 }
      "create"    { ::fwk::createProject }
      "build"     { ::fwk::buildProject }
      "postbuild" { ::fwk::postBuildProject }
      "simulate"  { ::fwk::simulateProject }
      "synth"     { ::fwk::synthProject }
      "clean"     { ::fwk::cleanProject $::fwk::PrjBuildName }
      "test"      { ::fwk::testProject }
      "gui"       { ::fwk::guiProject }
      "sdk"       { ::fwk::openSdk }
      default  {
        ::fwk::printError "Unrecognized command: $task"
        return -1
      }
    }
  }
  return 0
}

# rerun main init with all under tool shell
proc ::fwk::reinit {} {
  set currPath [pwd]
  cd $::fwk::ProjectPath
  if { [catch { eval ::fwk::main $::argv } resulttext] } {}
  cd $currPath
}

# wrap source procedure to add tool specific switches
proc sourceFwkFile { sourceFile } {
  if { $::fwk::GuiMode && ($::fwk::ToolType == "vivado" || $::fwk::ToolType == "planahead" )} {
    set script "source -notrace $sourceFile"
    uplevel $script
  } else {
    set script "source $sourceFile"
    uplevel $script
  }
}

proc ::fwk::exit { exitCode } {
  if { $::fwk::ExitMode == 1 } {
    ::exit $exitCode
  } else {
    ::fwk::printError "Exit code $exitCode"
    return -code error "Process Failed"
  }
}

proc ::fwk::procErrorInfo {errInfo} {
  set nameSpace ""
  set line 0
  set nsMatch [regexp {in namespace eval\s+"([a-zA-Z0-9_:]+)"\s+} $errInfo totalmach nameSpace]
  set match [regexp {procedure\s+"([a-zA-Z0-9_:]+)"\s+line\s*(\d+)} $errInfo totalmach procName line]
  set glbProc [regexp {^::([a-zA-Z0-9_:]+)} $procName]
  if { $match == 1 } {
    # puts ""
    if { $nsMatch == 1 && $glbProc == 0} {set procName ${nameSpace}::${procName}}
    puts "In procedure $procName at line $line"
    set lines [split [info body $procName] "\n"]
    set endLine [llength $lines]
    set startLine 0
    if { $line-4 >= 0 } { set startLine [expr {$line-4}] }
    if { $line+4 < $endLine } { set endLine [expr {$line+4}] }

    puts "..."
    for {set var $startLine} { $var < $endLine } { incr var } {
      set curLine [lindex $lines [expr {$var-1}]]
      if { $var == $line } {
        ::fwk::printc {-fg red -style bright} "> $curLine "
        #puts "> $curLine "
      } else { puts $curLine }
    }
    puts "..."
  }
  # add in file path info
  set match [regexp {file\s+"([a-zA-Z0-9_\ \.\/]+)"\s+line\s*(\d+)} $errInfo totalmach fileName lineNumber]
  if { $match == 1 } {
    puts "\nLine $lineNumber in file: $fileName:$lineNumber"
    puts "..."
    set fid [open $fileName r]
    set currLine 0
    while {1} {
      incr currLine
      if {[gets $fid line] == -1} break
      if {$currLine == $lineNumber} {
        ::fwk::printc {-fg red -style bright} "> $line"
      }
    }
    puts "..."
    close $fid
  }

}

proc ::fwk::catchHdl { command {message ""} } {
  if { [catch { uplevel $command } errmsg ] } {
    puts "\n"
    if {$message != "" } { ::fwk::printError $message }
    set einfo $::errorInfo
    set ecode $::errorCode
    # puts $einfo
    ::fwk::printError $errmsg
    ::fwk::procErrorInfo $einfo
    ::fwk::printError "Project script failed with error code: $ecode"
    puts ""
    ::fwk::exit -2
  }
}

# -----------------------------------------------------------------------------
# main run if not sourced
if { [info script] == $::argv0 } {
  if { [catch { eval ::fwk::main $::argv } errmsg] } {
    set einfo $::errorInfo
    set ecode $::errorCode
    puts "\n"
    ::fwk::printError $errmsg
    ::fwk::procErrorInfo $einfo
    ::fwk::printError "Project script failed with error code: $ecode"
    puts ""
    ::fwk::exit -10
  }
}

